/*
 * WSP BNF functions
 *
 * Version 1.0
 *
 * Olivier Biot
 */

/* These defs do not really fit in here but I need them */

static const value_string vals_header_names[] = {
    /* Encoding version 1.1 */
    {0x00, "Accept"},
    {0x01, "Accept-Charset"},
    {0x02, "Accept-Encoding"},
    {0x03, "Accept-Language"},
    {0x04, "Accept-Ranges"},
    {0x05, "Age"},
    {0x06, "Allow"},
    {0x07, "Authorization"},
    {0x08, "Cache-Control"},
    {0x09, "Connection"},
    {0x0A, "Content-Base"},
    {0x0B, "Content-Encoding"},
    {0x0C, "Content-Language"},
    {0x0D, "Content-Length"},
    {0x0E, "Content-Location"},
    {0x0F, "Content-MD5"},
    {0x10, "Content-Range"},
    {0x11, "Content-Type"},
    {0x12, "Date"},
    {0x13, "Etag"},
    {0x14, "Expires"},
    {0x15, "From"},
    {0x16, "Host"},
    {0x17, "If-Modified-Since"},
    {0x18, "If-Match"},
    {0x19, "If-None-Match"},
    {0x1A, "If-Range"},
    {0x1B, "If-Unmodified-Since"},
    {0x1C, "Location"},
    {0x1D, "Last-Modified"},
    {0x1E, "Max-Forwards"},
    {0x1F, "Pragma"},
    {0x20, "Proxy-Authenticate"},
    {0x21, "Proxy-Authorization"},
    {0x22, "Public"},
    {0x23, "Range"},
    {0x24, "Referer"},
    {0x25, "Retry-After"},
    {0x26, "Server"},
    {0x27, "Transfer-Encoding"},
    {0x28, "Upgrade"},
    {0x29, "User-Agent"},
    {0x2A, "Vary"},
    {0x2B, "Via"},
    {0x2C, "Warning"},
    {0x2D, "WWW_Authenticate"},
    {0x2E, "Content-Disposition"},
    /* Encoding version 1.2 */
    {0x2F, "X-Wap-Application-Id"},
    {0x30, "X-Wap-Content-URI"},
    {0x31, "X-Wap-Initiator-URI"},
    {0x32, "Accept-Application"},
    {0x33, "Bearer-Indication"},
    {0x34, "Push-Flag"},
    {0x35, "Profile"},
    {0x36, "Profile-Diff"},
    {0x37, "Profile-Warning"},
    /* Encoding version 1.3 */
    {0x38, "Expect"},
    {0x39, "TE"},
    {0x3A, "Trailer"},
    {0x3B, "Accept-Charset"},
    {0x3C, "Accept-Encoding"},
    {0x3D, "Cache-Control"},
    {0x3E, "Content-Range"},
    {0x3F, "X-Wap-Tod"},
    {0x40, "Content-Id"},
    {0x41, "Set-Cookie"},
    {0x42, "Cookie"},
    {0x43, "Encoding-Version"},

    {0x00, NULL}
};

#ifdef xyzzy
static const value_string vals_content_types[] = {
    /* Encoding version 1.1 */
    {0x00, "*/*"},
    {0x01, "text/*"},
    {0x02, "text/html"},
    {0x03, "text/plain"},
    {0x04, "text/x-hdml"},
    {0x05, "text/x-ttml"},
    {0x06, "text/x-vCalendar"},
    {0x07, "text/x-vCard"},
    {0x08, "text/vnd.wap.wml"},
    {0x09, "text/vnd.wap.wmlscript"},
    {0x0A, "text/vnd.wap.channel"},
    {0x0B, "Multipart/*"},
    {0x0C, "Multipart/mixed"},
    {0x0D, "Multipart/form-data"},
    {0x0E, "Multipart/byteranges"},
    {0x0F, "Multipart/alternative"},
    {0x10, "application/*"},
    {0x11, "application/java-vm"},
    {0x12, "application/x-www-form-urlencoded"},
    {0x13, "application/x-hdmlc"},
    {0x14, "application/vnd.wap.wmlc"},
    {0x15, "application/vnd.wap.wmlscriptc"},
    {0x16, "application/vnd.wap.channelc"},
    {0x17, "application/vnd.wap.uaprof"},
    {0x18, "application/vnd.wap.wtls-ca-certificate"},
    {0x19, "application/vnd.wap.wtls-user-certificate"},
    {0x1A, "application/x-x509-ca-cert"},
    {0x1B, "application/x-x509-user-cert"},
    {0x1C, "image/*"},
    {0x1D, "image/gif"},
    {0x1E, "image/jpeg"},
    {0x1F, "image/tiff"},
    {0x20, "image/png"},
    {0x21, "image/vnd.wap.wbmp"},
    {0x22, "application/vnd.wap.multipart.*"},
    {0x23, "application/vnd.wap.multipart.mixed"},
    {0x24, "application/vnd.wap.multipart.form-data"},
    {0x25, "application/vnd.wap.multipart.byteranges"},
    {0x26, "application/vnd.wap.multipart.alternative"},
    {0x27, "application/xml"},
    {0x28, "text/xml"},
    {0x29, "application/vnd.wap.wbxml"},
    {0x2A, "application/x-x968-cross-cert"},
    {0x2B, "application/x-x968-ca-cert"},
    {0x2C, "application/x-x968-user-cert"},
    {0x2D, "text/vnd.wap.si"},
    /* Encoding version 1.2 */
    {0x2E, "application/vnd.wap.sic"},
    {0x2F, "text/vnd.wap.sl"},
    {0x30, "application/vnd.wap.slc"},
    {0x31, "text/vnd.wap.co"},
    {0x32, "application/vnd.wap.coc"},
    {0x33, "application/vnd.wap.multipart.related"},
    {0x34, "application/vnd.wap.sia"},
    /* Encoding version 1.3 */
    {0x35, "text/vnd.wap.connectivity-xml"},
    {0x36, "application/vnd.wap.connectivity-wbxml"},
    {0x00, NULL}
};
#endif

/*
 * Basic BNF rules used in WSP, based on
 * RFC2616: "Hypertext Transfer Protocol -- HTTP/1.1"
 * ==================================================
 * 
 * OCTET		:= <Any 8-bits sequence of data>
 * CHAR			:= <Any US-ASCII character>
 * 					== Octets 0-127
 * CTL			:= <Any US-ASCII control character + DEL>
 * 					== Octets 0-31 + Octet 127
 * CR			:= <US-ASCII Carriage-return
 * 					== Octet 13
 * LF			:= <US-ASCII Line-feed
 * 					== Octet 10
 * SP			:= <US-ASCII Space>
 * 					== Octet 32
 * HT			:= <US-ASCII Horizontal-tab>
 * 					== Octet 9
 * <">			:= <US-ASCII Double-quote>
 * 					== Octet 34
 * 
 * CRLF			:= CR LF
 * LWS			:= [ CRLF ] 1*( SP | HT )
 * TEXT			:= <Any OCTET except CTL but including LWS>
 * 					== (Octets 32-126) | (Octets 128-255) | LWS
 * separators	:= "(" | ")" | "<" | ">" | "@"
 * 				 | "," | ";" | ":" | "\" | <">
 * 				 | "/" | "[" | "]" | "?" | "="
 * 				 | "{" | "}" | SP  | HT
 * 					== Octets 40 41 60 62 64 44 59 58 92 34 47 91 93 63 61 123 125 32 9
 * 					== Octets 9 32 34 40 41 44 47 58 59 60 61 62 63 64 91 92 93 123 125
 * token		:= 1*<Any CHAR except CTL or separators>
 * 					== 1*((Octet 32-126) - (Octets 9 32 34 40-41 44 47 58-64 91-93 123 125))
 *					== Octets 33 35-39 42-43 45-46 48-57 65-90 94-122 124 126
 */


/* All get_bnf_XXX() functions return a freshly malloc()'ed buffer.
 */
char *get_bnf_text_string		(tvbuff_t *tvb, int offset, guint *count) ;
char istokenchar				(int c) ;
char *get_bnf_token_text		(tvbuff_t *tvb, int offset, guint *count) ;
guint32 *get_bnf_short_integer	(tvbuff_t *tvb, int offset, guint *count) ;
guint32 *get_bnf_long_integer	(tvbuff_t *tvb, int offset, guint *count) ;
guint32 *get_bnf_integer_value	(tvbuff_t *tvb, int offset, guint *count) ;
guint32 *get_bnf_q_value		(tvbuff_t *tvb, int offset, guint *count) ;
char *get_bnf_version_value		(tvbuff_t *tvb, int offset, guint *count) ;

/* Functions */

char *get_bnf_text_string		(tvbuff_t *tvb, int offset, guint *count)
{
	guint	size	= 0,
			len		= tvb_reported_length (tvb);
	guint8	peek;
	int		i;
	char	*s;

	msg (MSG_INFO, "get_bnf_text_string(): offset = %d, len = %d, maxsize = %d", offset, len, len - offset);
	if (! tvb) {
		msg (MSG_WARN, "get_bnf_text_string(): tvb == NULL.");
		return NULL;
	}
	if (offset >= len) {	/* Buffer overrun */
		msg (MSG_ERROR, "get_bnf_text_string(): buffer overrun: offset = %d, len = %d", offset, len);
		return NULL;
	}

	/* text-string	= [quote] *TEXT end-of-string
	 * 				= [(0x7F)] *(0x20-0x7F) (0x00)
	 */
	
	i = offset;
	/* Get string length */
	while ( (peek = tvb_get_guint8(tvb, i)) ) {
		if (i >= len)
			break;
		if (peek >= 0x7F)
			size += 6; /* "<0x..>" for non-printable */
		else
			size++;
		i++;
	}
	(*count) += 1 + (i - offset);
	
	if (peek) { /* Stopped in middle of string */
		msg (MSG_ERROR, "get_bnf_text_string(): buffer overrun: offset = %d, len = %d", offset, len);
		return NULL;
	}
	s = malloc ((1 + size) * sizeof (char));
	if (s) {
		guint32	j	= 0;
		for (i = offset; i < offset + (*count); i++) {
			peek = tvb_get_guint8 (tvb, i);
			if (peek >= 0x7F) { /* "<0x..>" for non-printable */
				sprintf (s+j, "<0x%02X>", peek);
				j += 6;
			} else {
				s[j] = peek;
				j++;
			}
		}
		s[size] = 0;
		msg (MSG_INFO, "get_bnf_text_string(): offset = %d, len = %d, count = %d", offset, len, *count);
		return s;
	}
	msg (MSG_ERROR, "get_bnf_text_string(): malloc() failed.");
	return NULL;
}



char istokenchar (int c) {
	const char ch[] =
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x21, 0x00, 0x23, 0x24, 0x25, 0x26, 0x27, 0x00, 0x00, 0x2A, 0x2B, 0x00, 0x2D, 0x2E, 0x00,
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
		0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x00, 0x00, 0x00, 0x5E, 0x5F,
		0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
		0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x00, 0x7C, 0x00, 0x7E, 0x00,
	};
	if ((0 <= c) && (c <= 0x7F))
		return ch[c];
	return 0x00;
}



char *get_bnf_token_text		(tvbuff_t *tvb, int offset, guint *count)
{
	guint	len = tvb_reported_length (tvb);
	guint8	peek;
	char	*s;

	msg (MSG_INFO, "get_bnf_token_text(): offset = %d, len = %d, maxsize = %d", offset, len, len - offset);
	if (! tvb) {
		msg (MSG_WARN, "get_bnf_token_text(): tvb == NULL.");
		return NULL;
	}
	if (offset >= len) {	/* Buffer overrun */
		msg (MSG_ERROR, "get_bnf_token_text(): buffer overrun: offset = %d, len = %d", offset, len);
		return NULL;
	}
	
	peek	= tvb_get_guint8 (tvb, offset);
	if ( ! (istokenchar (peek)) ) { /* Invalid */
		msg (MSG_WARN, "get_bnf_token_text(): invalid token-text.");
		return NULL;
	}
	s = get_bnf_text_string (tvb, offset, count);
	msg (MSG_INFO, "get_bnf_token_text(): offset = %d, len = %d, count = %d", offset, len, *count);
	return s;
}



guint32 *get_bnf_short_integer	(tvbuff_t *tvb, int offset, guint *count)
{
	guint	len = tvb_reported_length (tvb);
	guint8	peek;
	guint32	*s;

	msg (MSG_INFO, "get_bnf_short_integer(): offset = %d, len = %d, maxsize = %d", offset, len, len - offset);
	if (! tvb) {
		msg (MSG_WARN, "get_bnf_short_integer(): tvb == NULL.");
		return NULL;
	}
	if (offset >= len) {	/* Buffer overrun */
		msg (MSG_ERROR, "get_bnf_short_integer(): buffer overrun: offset = %d, len = %d", offset, len);
		return NULL;
	}

	peek	= tvb_get_guint8 (tvb, offset);

	if (peek & 0x80) { /* Short-integer := <Octet 0x80-0xFF> */
		(*count)++;
		s = malloc (sizeof (guint32));
		if (s) {
			(*s) = peek & 0x7F;
			msg (MSG_INFO, "get_bnf_short_integer(): offset = %d, len = %d, count = %d, value = %d", offset, len, *count, *s);
			return s;
		} else { /* Error: malloc() failed */ }
	} else { /* Invalid */
		msg (MSG_WARN, "get_bnf_integer_value(): invalid short-integer value.");
		return NULL;
	}
	msg (MSG_ERROR, "get_bnf_short_integer(): malloc() failed.");
	return NULL;
}



guint32 *get_bnf_long_integer	(tvbuff_t *tvb, int offset, guint *count)
{
	guint	len = tvb_reported_length (tvb);
	guint8	peek;
	guint32	*s;

	msg (MSG_INFO, "get_bnf_long_integer(): offset = %d, len = %d, maxsize = %d", offset, len, len - offset);
	if (! tvb) {
		msg (MSG_WARN, "get_bnf_long_integer(): tvb == NULL.");
		return NULL;
	}
	if (offset >= len) {	/* Buffer overrun */
		msg (MSG_ERROR, "get_bnf_long_integer(): buffer overrun: offset = %d, len = %d", offset, len);
		return NULL;
	}
	
	/* Long-integer  := Short-length Multi-octet-integer
	 * Short-Length  := <Octet 0x00-0x1E>
	 */
	peek	= tvb_get_guint8 (tvb, offset);
	msg (MSG_INFO, "get_bnf_long_integer(): first octet is 0x%02X.", peek);
	(*count)++;
	offset++;

	if (peek < 0x1F) { /* Short-length */
		if (offset + peek > len) { /* Buffer overrun */
			msg (MSG_ERROR, "get_bnf_long_integer(): buffer overrun: offset + peek = %d, len = %d", offset + peek, len);
			return NULL;
		}
		(*count) += peek;
		msg (MSG_INFO, "get_bnf_long_integer(): offset = %d, len = %d, count = %d", offset, len, *count);
		if (peek < 5) { /* max 32 bits */
			guint32	val	= 0;
			char	i	= 0;
			while (i < peek) {
				val = (val << 8) + tvb_get_guint8(tvb, offset + i);
				i++;
			}
			s = malloc (sizeof (guint32));
			if (s) {
				(*s) = val;
				return s;
			}
		} else { /* bigger than 32 bits: stays hex */
				msg (MSG_WARN, "get_bnf_long_integer(): integer too big for 32-bit assumption (len = %d, count = %d)", len, *count);
				return NULL;
		}
	} else {
		msg (MSG_WARN, "get_bnf_long_integer(): invalid long-integer.");
		return NULL;
	}
	msg (MSG_ERROR, "get_bnf_long_integer(): malloc() failed.");
	return NULL;
}



guint32 *get_bnf_integer_value	(tvbuff_t *tvb, int offset, guint *count)
{
	guint	len		= tvb_reported_length (tvb);
	guint8	peek;
	guint32	*s;

	msg (MSG_INFO, "get_bnf_integer_value(): offset = %d, len = %d, maxsize = %d", offset, len, len - offset);
	if (! tvb) {
		msg (MSG_WARN, "get_bnf_integer_value(): tvb == NULL.");
		return NULL;
	}
	if (offset >= len) {	/* Buffer overrun */
		msg (MSG_ERROR, "get_bnf_integer_value(): buffer overrun: offset = %d, len = %d", offset, len);
		return NULL;
	}
	/* Integer-value := Short-integer | Long-integer
	 * Short-integer := <Octet >= 0x80>
	 * Long-integer  := Short-length Multi-octet-integer
	 * Short-Length  := <Octet 0x00-0x1E>
	 */
	peek	= tvb_get_guint8 (tvb, offset);
	msg (MSG_INFO, "get_bnf_integer_value(): first octet is 0x%02X.", peek);

	if (peek & 0x80) { /* Short-integer */
		s = get_bnf_short_integer (tvb, offset, count);
		msg (MSG_INFO, "get_bnf_integer_value(): offset = %d, len = %d, count = %d, value = %d", offset, len, *count, *s);
		if (s)
			return s;
	} else if (peek < 0x1F) { /* Long-integer */
		s = get_bnf_long_integer (tvb, offset, count);
		msg (MSG_INFO, "get_bnf_integer_value(): offset = %d, len = %d, count = %d", offset, len, *count);
		if (s)
			return s;
	} else { /* Invalid */
		msg (MSG_WARN, "get_bnf_integer_value(): invalid integer-value.");
		return NULL;
	}
	msg (MSG_ERROR, "get_bnf_integer_value(): malloc() failed.");
	return NULL;
}



guint32 *get_bnf_q_value		(tvbuff_t *tvb, int offset, guint *count)
{
	guint	len = tvb_reported_length (tvb);
	guint8	peek;
	guint32	*s;

	msg (MSG_INFO, "get_bnf_q_value(): offset = %d, len = %d, maxsize = %d", offset, len, len - offset);
	if (! tvb) {
		msg (MSG_WARN, "get_bnf_q_value(): tvb == NULL.");
		return NULL;
	}
	if (offset >= len) {	/* Buffer overrun */
		msg (MSG_ERROR, "get_bnf_q_value(): buffer overrun: offset = %d, len = %d", offset, len);
		return NULL;
	}
	peek	= tvb_get_guint8 (tvb, offset);
	offset++;
	(*count)++;
	/* Encoding of Q-value:
	 * 0x01   - 0x64   : 0.00  - 0.99
	 * 0x65   - 0x7F   : 0.001 - 0.027
	 * 0x8100 - 0x8371 : 0.028 - 0.999
	 */
	if (peek < 0x65) {
		s = malloc (6 * sizeof (guint32));
		if (s) {
			(*s) = 10 * (peek - 1);
			msg (MSG_INFO, "get_bnf_q_value(): offset = %d, len = %d, count = %d", offset, len, *count);
			return s;
		}
	} else if (peek < 0x80) {
		s = malloc (6 * sizeof (guint32));
		if (s) {
			(*s) = (peek - 100);
			msg (MSG_INFO, "get_bnf_q_value(): offset = %d, len = %d, count = %d", offset, len, *count);
			return s;
		}
	} else { /* Retrieve next byte */
		guint8	peek2;

		if (offset >= len) {	/* Buffer overrun */
			msg (MSG_ERROR, "get_bnf_q_value(): buffer overrun: offset = %d, len = %d", offset, len);
			return NULL;
		}
		peek2	= tvb_get_guint8 (tvb, offset);
		(*count)++;
		if (peek2 & 0x80) { /* Error: invalid Q format */
			msg (MSG_WARN, "get_bnf_q_value(): invalid Q-value.");
			return NULL;
		} else {
			guint32	val = ((peek & 0x7F) << 7) + peek2 - 100;
			if (val < 1000) { /* OK */
				s = malloc (sizeof (guint32));
				if (s) {
					(*s) = val;
					msg (MSG_INFO, "get_bnf_q_value(): offset = %d, len = %d, count = %d", offset, len, *count);
					return s;
				}
			} else { /* Invalid Q-value */
				msg (MSG_WARN, "get_bnf_q_value(): invalid Q-value.");
				return NULL;
			}
		}
	}
	msg (MSG_ERROR, "get_bnf_q_value(): malloc() failed.");
	return NULL;
}



char *get_bnf_version_value		(tvbuff_t *tvb, int offset, guint *count)
{
	guint	len = tvb_reported_length (tvb);
	guint8	peek,
			minor,
			major;
	char	*s;

	msg (MSG_INFO, "get_bnf_version_value(): offset = %d, len = %d, maxsize = %d", offset, len, len - offset);
	if (! tvb) {
		msg (MSG_WARN, "get_bnf_version_value(): tvb == NULL.");
		return NULL;
	}
	if (offset >= len) {	/* Buffer overrun */
		msg (MSG_ERROR, "get_bnf_version_value(): buffer overrun: offset = %d, len = %d", offset, len);
		return NULL;
	}
	/* Version-value := short-integer | text-string
	 * For short-integer:
	 * 		Version-major := (b & 0x70) >> 4;
	 * 		Version-minor := (b & 0x0F);
	 * 		IF (Version-minor== 0x0F) THEN no_version_minor
	 */
	peek	= tvb_get_guint8 (tvb, offset);
	if (peek & 0x80) { /* Short-integer */
		(*count)++;
		major	= (peek & 0x70) >> 4;
		minor	= (peek & 0x0F);
		s = malloc (5 * sizeof (char));
		if (s) {
			if (minor == 0x0F) {
				sprintf(s, "%1d", major);
			} else {
				sprintf(s, "%1d.%2d", major, minor);
			}
			msg (MSG_INFO, "get_bnf_version_value(): offset = %d, len = %d, count = %d", offset, len, *count);
			return s;
		}
	} else { /* Text-string */
		s = get_bnf_text_string (tvb, offset, count);
		msg (MSG_INFO, "get_bnf_version_value(): offset = %d, len = %d, count = %d", offset, len, *count);
		if (s)
			return s;
	}
	msg (MSG_ERROR, "get_bnf_version_value(): malloc() failed.");
	return NULL;
}

/* wsp-bnf.c */
