/* packet-wsp.c
 *
 * Routines to dissect WSP component of WAP traffic.
 * 
 * $Id: packet-wsp.c,v 1.19 2001/02/19 21:02:33 guy Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@zing.org>
 * Copyright 1998 Didier Jorand
 *
 * WAP dissector based on original work by Ben Fowler
 * Updated by Neil Hunter <neil.hunter@energis-squared.com>
 * WTLS support by Alexandre P. Ferreira (Splice IP)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#ifdef NEED_SNPRINTF_H
# ifdef HAVE_STDARG_H
#  include <stdarg.h>
# else
#  include <varargs.h>
# endif
# include "snprintf.h"
#endif

#include <string.h>
#include <glib.h>
#include <ctype.h>

#include "packet.h"
#include "packet-wap.h"
#include "packet-wsp.h"

/* File scoped variables for the protocol and registered fields */
static int proto_wsp 							= HF_EMPTY;

/* These fields used by fixed part of header */
static int hf_wsp_header_tid					= HF_EMPTY;
static int hf_wsp_header_pdu_type				= HF_EMPTY;
static int hf_wsp_version_major					= HF_EMPTY;
static int hf_wsp_version_minor					= HF_EMPTY;
static int hf_wsp_capability_length				= HF_EMPTY;
static int hf_wsp_capabilities_section			= HF_EMPTY;
static int hf_wsp_header_uri_len				= HF_EMPTY;
static int hf_wsp_header_uri					= HF_EMPTY;
static int hf_wsp_server_session_id				= HF_EMPTY;
static int hf_wsp_header_status					= HF_EMPTY;
static int hf_wsp_header_length					= HF_EMPTY;
static int hf_wsp_headers_section				= HF_EMPTY;
static int hf_wsp_header						= HF_EMPTY;
static int hf_wsp_content_type					= HF_EMPTY;
static int hf_wsp_parameter_well_known_charset	= HF_EMPTY;
static int hf_wsp_reply_data					= HF_EMPTY;
static int hf_wsp_post_data						= HF_EMPTY;

static int hf_wsp_header_accept					= HF_EMPTY;
static int hf_wsp_header_accept_str				= HF_EMPTY;
static int hf_wsp_header_accept_charset			= HF_EMPTY;
static int hf_wsp_header_accept_charset_str		= HF_EMPTY;
static int hf_wsp_header_accept_language		= HF_EMPTY;
static int hf_wsp_header_accept_language_str	= HF_EMPTY;
static int hf_wsp_header_content_language		= HF_EMPTY;
static int hf_wsp_header_content_language_str	= HF_EMPTY;
static int hf_wsp_header_accept_ranges			= HF_EMPTY;
static int hf_wsp_header_accept_ranges_str		= HF_EMPTY;
static int hf_wsp_header_cache_control				= HF_EMPTY;
static int hf_wsp_header_cache_control_tok			= HF_EMPTY;
static int hf_wsp_header_cache_control_time			= HF_EMPTY;
static int hf_wsp_header_cache_control_header		= HF_EMPTY;
static int hf_wsp_header_cache_control_header_tok	= HF_EMPTY;

static int hf_wsp_header_cache_control_max_age		= HF_EMPTY;
static int hf_wsp_header_cache_control_max_stale	= HF_EMPTY;
static int hf_wsp_header_cache_control_min_fresh	= HF_EMPTY;
static int hf_wsp_header_cache_control_no_cache		= HF_EMPTY;
static int hf_wsp_header_cache_control_no_cache_str	= HF_EMPTY;
static int hf_wsp_header_cache_control_private		= HF_EMPTY;
static int hf_wsp_header_cache_control_private_str	= HF_EMPTY;

static int hf_wsp_header_content_length			= HF_EMPTY;
static int hf_wsp_header_age					= HF_EMPTY;
static int hf_wsp_header_date					= HF_EMPTY;
static int hf_wsp_header_etag					= HF_EMPTY;
static int hf_wsp_header_expires				= HF_EMPTY;
static int hf_wsp_header_last_modified			= HF_EMPTY;
static int hf_wsp_header_location				= HF_EMPTY;
static int hf_wsp_header_if_modified_since		= HF_EMPTY;
static int hf_wsp_header_server					= HF_EMPTY;
static int hf_wsp_header_user_agent				= HF_EMPTY;
static int hf_wsp_header_application_header		= HF_EMPTY;
static int hf_wsp_header_application_value		= HF_EMPTY;
static int hf_wsp_header_x_wap_tod				= HF_EMPTY;
static int hf_wsp_header_transfer_encoding		= HF_EMPTY;
static int hf_wsp_header_transfer_encoding_str	= HF_EMPTY;
static int hf_wsp_header_via					= HF_EMPTY;
/* (quoted) text string */
static int hf_wsp_header_from					= HF_EMPTY;
static int hf_wsp_header_host					= HF_EMPTY;
static int hf_wsp_header_if_match				= HF_EMPTY;
static int hf_wsp_header_if_none_match			= HF_EMPTY;
static int hf_wsp_header_upgrade				= HF_EMPTY;
static int hf_wsp_header_content_id				= HF_EMPTY;
/* URI value */
static int hf_wsp_header_content_base			= HF_EMPTY;
static int hf_wsp_header_content_location		= HF_EMPTY;
static int hf_wsp_header_referer				= HF_EMPTY;
static int hf_wsp_header_x_wap_content_uri		= HF_EMPTY;
static int hf_wsp_header_x_wap_initiator_uri	= HF_EMPTY;
static int hf_wsp_header_profile				= HF_EMPTY;
static int hf_wsp_header_if_unmodified_since	= HF_EMPTY;
static int hf_wsp_header_if_range_str			= HF_EMPTY;
static int hf_wsp_header_if_range_date			= HF_EMPTY;
/* various similar to Accept */
static int hf_wsp_header_accept_encoding		= HF_EMPTY;
static int hf_wsp_header_accept_encoding_str	= HF_EMPTY;
static int hf_wsp_header_content_encoding		= HF_EMPTY;
static int hf_wsp_header_content_encoding_str	= HF_EMPTY;
static int hf_wsp_header_warning				= HF_EMPTY;
static int hf_wsp_header_warning_str			= HF_EMPTY;
/* Short-int */
static int hf_wsp_header_allow					= HF_EMPTY;
static int hf_wsp_header_push_flag				= HF_EMPTY;
/* Short-int | Token-text */
static int hf_wsp_header_connection				= HF_EMPTY;
static int hf_wsp_header_public					= HF_EMPTY;
static int hf_wsp_header_public_str				= HF_EMPTY;
static int hf_wsp_header_vary					= HF_EMPTY;
static int hf_wsp_header_vary_str				= HF_EMPTY;
/* Integer-value | Token-text */
static int hf_wsp_header_trailer				= HF_EMPTY;
static int hf_wsp_header_trailer_str			= HF_EMPTY;
/* Short-int | Parameter */
static int hf_wsp_header_pragma					= HF_EMPTY;
/* Integer-value */
static int hf_wsp_header_max_forwards			= HF_EMPTY;
static int hf_wsp_header_bearer_indication		= HF_EMPTY;
/* profile-warning */
static int hf_wsp_header_profile_warning_code	= HF_EMPTY;
static int hf_wsp_header_profile_warning_uri	= HF_EMPTY;
static int hf_wsp_header_profile_warning_date	= HF_EMPTY;
/* Expect */
static int hf_wsp_header_expect_text			= HF_EMPTY;
static int hf_wsp_header_expect_parm			= HF_EMPTY;
/* Authoization */
static int hf_wsp_header_authorization			= HF_EMPTY;
static int hf_wsp_header_proxy_authorization	= HF_EMPTY;
static int hf_wsp_header_proxy_authenticate		= HF_EMPTY;
static int hf_wsp_header_www_authenticate		= HF_EMPTY;
/* WSP redirect PDU */
static int hf_wsp_redirect_flags				= HF_EMPTY;
static int hf_wsp_redirect_addresses_section	= HF_EMPTY;
static int hf_wsp_redirect_address_section		= HF_EMPTY;
static int hf_wsp_redirect_address_length		= HF_EMPTY;
static int hf_wsp_redirect_port_included		= HF_EMPTY;
static int hf_wsp_redirect_port					= HF_EMPTY;
static int hf_wsp_redirect_bearertype_included	= HF_EMPTY;
static int hf_wsp_redirect_bearertype			= HF_EMPTY;
/* WSP capabilities */
static int hf_wsp_capability_section				= HF_EMPTY;
static int hf_wsp_capability_client_sdu_size		= HF_EMPTY;
static int hf_wsp_capability_server_sdu_size		= HF_EMPTY;
static int hf_wsp_capability_proto_options			= HF_EMPTY;
static int hf_wsp_capability_proto_option			= HF_EMPTY;
static int hf_wsp_capability_method_mor				= HF_EMPTY;
static int hf_wsp_capability_push_mor				= HF_EMPTY;
static int hf_wsp_capability_xmethods				= HF_EMPTY;
static int hf_wsp_capability_xmethod_get			= HF_EMPTY;
static int hf_wsp_capability_xmethod_get_val		= HF_EMPTY;
static int hf_wsp_capability_xmethod_push			= HF_EMPTY;
static int hf_wsp_capability_xmethod_push_val		= HF_EMPTY;
static int hf_wsp_capability_header_code_pages		= HF_EMPTY;
static int hf_wsp_capability_header_code_page		= HF_EMPTY;
static int hf_wsp_capability_header_code_page_val	= HF_EMPTY;
static int hf_wsp_capability_alias_address_section		= HF_EMPTY;
static int hf_wsp_capability_alias_address_length		= HF_EMPTY;
static int hf_wsp_capability_alias_port_included		= HF_EMPTY;
static int hf_wsp_capability_alias_port					= HF_EMPTY;
static int hf_wsp_capability_alias_bearertype_included	= HF_EMPTY;
static int hf_wsp_capability_alias_bearertype			= HF_EMPTY;

/* Initialize the subtree pointers */
static gint ett_wsp 							= ETT_EMPTY;
static gint ett_header 							= ETT_EMPTY;
static gint ett_headers							= ETT_EMPTY;
static gint ett_capabilities					= ETT_EMPTY;
static gint ett_content_type					= ETT_EMPTY;
static gint ett_redirect						= ETT_EMPTY;
static gint ett_redir_addr						= ETT_EMPTY;

/* Handle for WMLC dissector */
static dissector_handle_t wmlc_handle;

static const value_string vals_pdu_type[] = {
	{ 0x00, "Reserved" },
	{ 0x01, "Connect" },
	{ 0x02, "ConnectReply" },
	{ 0x03, "Redirect" },
	{ 0x04, "Reply" },
	{ 0x05, "Disconnect" },
	{ 0x06, "Push" },
	{ 0x07, "ConfirmedPush" },
	{ 0x08, "Suspend" },
	{ 0x09, "Resume" },

	/* 0x10 - 0x3F Unassigned */

	{ 0x40, "Get" },
	{ 0x41, "Options" },
	{ 0x42, "Head" },
	{ 0x43, "Delete" },
	{ 0x44, "Trace" },

	/* 0x45 - 0x4F Unassigned (Get PDU) */
	/* 0x50 - 0x5F Extended method (Get PDU) */
	{ 0x50, "Get: Extended method 01" },
	{ 0x51, "Get: Extended method 02" },
	{ 0x52, "Get: Extended method 03" },
	{ 0x53, "Get: Extended method 04" },
	{ 0x54, "Get: Extended method 05" },
	{ 0x55, "Get: Extended method 06" },
	{ 0x56, "Get: Extended method 07" },
	{ 0x57, "Get: Extended method 08" },
	{ 0x58, "Get: Extended method 09" },
	{ 0x59, "Get: Extended method 10" },
	{ 0x5A, "Get: Extended method 11" },
	{ 0x5B, "Get: Extended method 12" },
	{ 0x5C, "Get: Extended method 13" },
	{ 0x5D, "Get: Extended method 14" },
	{ 0x5E, "Get: Extended method 15" },
	{ 0x5F, "Get: Extended method 16" },

	{ 0x60, "Post" },
	{ 0x61, "Put" },

	/* 0x62 - 0x6F Unassigned (Post PDU) */
	/* 0x70 - 0x7F Extended method (Post PDU) */
	{ 0x70, "Post: Extended method 01" },
	{ 0x71, "Post: Extended method 02" },
	{ 0x72, "Post: Extended method 03" },
	{ 0x73, "Post: Extended method 04" },
	{ 0x74, "Post: Extended method 05" },
	{ 0x75, "Post: Extended method 06" },
	{ 0x76, "Post: Extended method 07" },
	{ 0x77, "Post: Extended method 08" },
	{ 0x78, "Post: Extended method 09" },
	{ 0x79, "Post: Extended method 10" },
	{ 0x7A, "Post: Extended method 11" },
	{ 0x7B, "Post: Extended method 12" },
	{ 0x7C, "Post: Extended method 13" },
	{ 0x7D, "Post: Extended method 14" },
	{ 0x7E, "Post: Extended method 15" },
	{ 0x7F, "Post: Extended method 16" },
	/* 0x80 - 0xFF Reserved */

	{ 0x00, NULL }

};

static const value_string vals_status[] = {
	/* 0x00 - 0x0F Reserved */

	{ 0x10, "100 Continue" },
	{ 0x11, "101 Switching Protocols" },

	{ 0x20, "200 OK" },
	{ 0x21, "201 Created" },
	{ 0x22, "202 Accepted" },
	{ 0x23, "203 Non-Authoritative Information" },
	{ 0x24, "204 No Content" },
	{ 0x25, "205 Reset Content" },
	{ 0x26, "206 Partial Content" },

	{ 0x30, "300 Multiple Choices" },
	{ 0x31, "301 Moved Permanently" },
	{ 0x32, "302 Moved Temporarily" },
	{ 0x33, "303 See Other" },
	{ 0x34, "304 Not Modified" },
	{ 0x35, "305 Use Proxy" },
	
	{ 0x37, "307 Temporary Redirect" },

	{ 0x40, "400 Bad Request" },
	{ 0x41, "401 Unauthorised" },
	{ 0x42, "402 Payment Required" },
	{ 0x43, "403 Forbidden" },
	{ 0x44, "404 Not Found" },
	{ 0x45, "405 Method Not Allowed" },
	{ 0x46, "406 Not Acceptable" },
	{ 0x47, "407 Proxy Authentication Required" },
	{ 0x48, "408 Request Timeout" },
	{ 0x49, "409 Conflict" },
	{ 0x4A, "410 Gone" },
	{ 0x4B, "411 Length Required" },
	{ 0x4C, "412 Precondition Failed" },
	{ 0x4D, "413 Request Entity Too Large" },
	{ 0x4E, "414 Request-URI Too Large" },
	{ 0x4F, "415 Unsupported Media Type" },
	{ 0x50, "416 Requested Range Not Satisfiable" },
	{ 0x51, "417 Expectation Failed" },

	{ 0x60, "500 Internal Server Error" },
	{ 0x61, "501 Not Implemented" },
	{ 0x62, "502 Bad Gateway" },
	{ 0x63, "503 Service Unavailable" },
	{ 0x64, "504 Gateway Timeout" },
	{ 0x65, "505 HTTP Version Not Supported" },
	
	{ 0x00, NULL }
};

/* Defined in http://www.wapforum.org/wina/wsp-content-type.htm */
static const value_string vals_content_types[] = {
	{ 0x00, "*/*" },
	{ 0x01, "text/*" },
	{ 0x02, "text/html" },
	{ 0x03, "text/plain" },
	{ 0x04, "text/x-hdml" },
	{ 0x05, "text/x-ttml" },
	{ 0x06, "text/x-vCalendar" },
	{ 0x07, "text/x-vCard" },
	{ 0x08, "text/vnd.wap.wml" },
	{ 0x09, "text/vnd.wap.wmlscript" },
	{ 0x0A, "text/vnd.wap.channel" },
	{ 0x0B, "Multipart/*" },
	{ 0x0C, "Multipart/mixed" },
	{ 0x0D, "Multipart/form-data" },
	{ 0x0E, "Multipart/byteranges" },
	{ 0x0F, "Multipart/alternative" },
	{ 0x10, "application/*" },
	{ 0x11, "application/java-vm" },
	{ 0x12, "application/x-www-form-urlencoded" },
	{ 0x13, "application/x-hdmlc" },
	{ 0x14, "application/vnd.wap.wmlc" },
	{ 0x15, "application/vnd.wap.wmlscriptc" },
	{ 0x16, "application/vnd.wap.channelc" },
	{ 0x17, "application/vnd.wap.uaprof" },
	{ 0x18, "application/vnd.wap.wtls-ca-certificate" },
	{ 0x19, "application/vnd.wap.wtls-user-certificate" },
	{ 0x1A, "application/x-x509-ca-cert" },
	{ 0x1B, "application/x-x509-user-cert" },
	{ 0x1C, "image/*" },
	{ 0x1D, "image/gif" },
	{ 0x1E, "image/jpeg" },
	{ 0x1F, "image/tiff" },
	{ 0x20, "image/png" },
	{ 0x21, "image/vnd.wap.wbmp" },
	{ 0x22, "application/vnd.wap.multipart.*" },
	{ 0x23, "application/vnd.wap.multipart.mixed" },
	{ 0x24, "application/vnd.wap.multipart.form-data" },
	{ 0x25, "application/vnd.wap.multipart.byteranges" },
	{ 0x26, "application/vnd.wap.multipart.alternative" },
	{ 0x27, "application/xml" },
	{ 0x28, "text/xml" },
	{ 0x29, "application/vnd.wap.wbxml" },
	{ 0x2A, "application/x-x968-cross-cert" },
	{ 0x2B, "application/x-x968-ca-cert" },
	{ 0x2C, "application/x-x968-user-cert" },
	{ 0x2D, "text/vnd.wap.si" },
	{ 0x2E, "application/vnd.wap.sic" },
	{ 0x2F, "text/vnd.wap.sl" },
	{ 0x30, "application/vnd.wap.slc" },
	{ 0x31, "text/vnd.wap.co" },
	{ 0x32, "application/vnd.wap.coc" },
	{ 0x33, "application/vnd.wap.multipart.related" },
	{ 0x34, "application/vnd.wap.sia" },

	{ 0x0201, "application/vnd.uplanet.cacheop-wbxml" },
	{ 0x0202, "application/vnd.uplanet.signal" },
	{ 0x0203, "application/vnd.uplanet.alert-wbxml" },
	{ 0x0204, "application/vnd.uplanet.list-wbxml" },
	{ 0x0205, "application/vnd.uplanet.listcmd-wbxml" },
	{ 0x0206, "application/vnd.uplanet.channel-wbxml" },
	{ 0x0207, "application/vnd.uplanet.provisioning-status-uri" },
	{ 0x0208, "x-wap.multipart/vnd.uplanet.header-set" },
	{ 0x0209, "application/vnd.uplanet.bearer-choice-wbxml" },
	{ 0x020A, "application/vnd.phonecom.mmc-wbxml" },

	{ 0x00, NULL }
};

static const value_string vals_languages[] = {
	{ 0x01, "Afar (aa)" },
	{ 0x02, "Abkhazian (ab)" },
	{ 0x03, "Afrikaans (af)" },
	{ 0x04, "Amharic (am)" },
	{ 0x05, "Arabic (ar)" },
	{ 0x06, "Assamese (as)" },
	{ 0x07, "Aymara (ay)" },
	{ 0x08, "Azerbaijani (az)" },
	{ 0x09, "Bashkir (ba)" },
	{ 0x0A, "Byelorussian (be)" },
	{ 0x0B, "Bulgarian (bg)" },
	{ 0x0C, "Bihari (bh)" },
	{ 0x0D, "Bislama (bi)" },
	{ 0x0E, "Bengali; Bangla (bn)" },
	{ 0x0F, "Tibetan (bo)" },
	{ 0x10, "Breton (br)" },
	{ 0x11, "Catalan (ca)" },
	{ 0x12, "Corsican (co)" },
	{ 0x13, "Czech (cs)" },
	{ 0x14, "Welsh (cy)" },
	{ 0x15, "Danish (da)" },
	{ 0x16, "German (de)" },
	{ 0x17, "Bhutani (dz)" },
	{ 0x18, "Greek (el)" },
	{ 0x19, "English (en)" },
	{ 0x1A, "Esperanto (eo)" },
	{ 0x1B, "Spanish (es)" },
	{ 0x1C, "Estonian (et)" },
	{ 0x1D, "Basque (eu)" },
	{ 0x1E, "Persian (fa)" },
	{ 0x1F, "Finnish (fi)" },
	{ 0x20, "Fiji (fj)" },
	{ 0x22, "French (fr)" },
	{ 0x24, "Irish (ga)" },
	{ 0x25, "Scots Gaelic (gd)" },
	{ 0x26, "Galician (gl)" },
	{ 0x27, "Guarani (gn)" },
	{ 0x28, "Gujarati (gu)" },
	{ 0x29, "Hausa (ha)" },
	{ 0x2A, "Hebrew (formerly iw) (he)" },
	{ 0x2B, "Hindi (hi)" },
	{ 0x2C, "Croatian (hr)" },
	{ 0x2D, "Hungarian (hu)" },
	{ 0x2E, "Armenian (hy)" },
	{ 0x30, "Indonesian (formerly in) (id)" },
	{ 0x47, "Maori (mi)" },
	{ 0x48, "Macedonian (mk)" },
	{ 0x49, "Malayalam (ml)" },
	{ 0x4A, "Mongolian (mn)" },
	{ 0x4B, "Moldavian (mo)" },
	{ 0x4C, "Marathi (mr)" },
	{ 0x4D, "Malay (ms)" },
	{ 0x4E, "Maltese (mt)" },
	{ 0x4F, "Burmese (my)" },
	{ 0x51, "Nepali (ne)" },
	{ 0x52, "Dutch (nl)" },
	{ 0x53, "Norwegian (no)" },
	{ 0x54, "Occitan (oc)" },
	{ 0x55, "(Afan) Oromo (om)" },
	{ 0x56, "Oriya (or)" },
	{ 0x57, "Punjabi (pa)" },
	{ 0x58, "Polish (po)" },
	{ 0x59, "Pashto, Pushto (ps)" },
	{ 0x5A, "Portuguese (pt)" },
	{ 0x5B, "Quechua (qu)" },
	{ 0x5D, "Kirundi (rn)" },
	{ 0x5E, "Romanian (ro)" },
	{ 0x5F, "Russian (ru)" },
	{ 0x60, "Kinyarwanda (rw)" },
	{ 0x61, "Sanskrit (sa)" },
	{ 0x62, "Sindhi (sd)" },
	{ 0x63, "Sangho (sg)" },
	{ 0x64, "Serbo-Croatian (sh)" },
	{ 0x65, "Sinhalese (si)" },
	{ 0x66, "Slovak (sk)" },
	{ 0x67, "Slovenian (sl)" },
	{ 0x68, "Samoan (sm)" },
	{ 0x69, "Shona (sn)" },
	{ 0x6A, "Somali (so)" },
	{ 0x6B, "Albanian (sq)" },
	{ 0x6C, "Serbian (sr)" },
	{ 0x6D, "Siswati (ss)" },
	{ 0x6E, "Sesotho (st)" },
	{ 0x6F, "Sundanese (su)" },
	{ 0x70, "Swedish (sv)" },
	{ 0x71, "Swahili (sw)" },
	{ 0x72, "Tamil (ta)" },
	{ 0x73, "Telugu (te)" },
	{ 0x74, "Tajik (tg)" },
	{ 0x75, "Thai (th)" },
	{ 0x76, "Tigrinya (ti)" },
	{ 0x81, "Nauru (na)" },
	{ 0x82, "Faeroese (fo)" },
	{ 0x83, "Frisian (fy)" },
	{ 0x84, "Interlingua (ia)" },
	{ 0x8C, "Rhaeto-Romance (rm)" },
	{ 0x00, NULL }
};

static const value_string vals_accept_ranges[] = {
	{ 0x80, "None" },
	{ 0x81, "Bytes" },
	{ 0x00, NULL }
};

static const value_string vals_cache_control[] = {
	{ 0x80, "No-cache" },
	{ 0x81, "No-store" },
	{ 0x82, "Max-age" },
	{ 0x83, "Max-stale" },
	{ 0x84, "Min-fresh" },
	{ 0x85, "Only-if-cached" },
	{ 0x86, "Public" },
	{ 0x87, "Private" },
	{ 0x88, "No-transform" },
	{ 0x89, "Must-revalidate" },
	{ 0x8A, "Proxy-revalidate" },
	{ 0x00, NULL }
};

static const value_string vals_transfer_encoding[] = {
	{ 0x80, "Chunked" },
	{ 0x00, NULL }
};

static const value_string vals_accept_encodings[] = {
	{ 0x80, "gzip" },
	{ 0x81, "compress" },
	{ 0x82, "deflate" },
	{ 0x83, "*" },
	{ 0x00, NULL }
};

static const value_string vals_content_encodings[] = {
	{ 0x80, "gzip" },
	{ 0x81, "compress" },
	{ 0x82, "deflate" },

	{ 0x00, NULL },
};

static const value_string vals_wsp_redirect_flags[] = {
	{ 0x00, "None" },
	{ 0x40, "Reuse security session"},
	{ 0x80, "Permanent redirect"},
	{ 0xC0, "Permanent redirect, reuse security session"},
	{ 0x00, NULL },
};

static const value_string vals_wsp_capability_proto_options[] = {
	{ 0x10, "Acknowledgement Headers" },
	{ 0x20, "Session Resume facility" },
	{ 0x40, "Push facility" },
	{ 0x80, "Confirmed Push facility" },
	{ 0x00, NULL },
};


/* See [WSP] and http://www.w3.org/TR/NOTE-CCPPexchange */
static const value_string vals_profile_warning_codes[] = {
	{ 0x10, "100 OK" },
	{ 0x11, "101 Used stale profile" },
	{ 0x12, "102 Not used profile" },
	{ 0x20, "200 Not applied" },
	{ 0x21, "201 Content selection applied" },
	{ 0x22, "202 Content generation applied" },
	{ 0x23, "203 Transformation applied" },

	{ 0x00, NULL },
};

/*
 * Windows appears to define DELETE.
 */
#ifdef DELETE
#undef DELETE
#endif

enum {
	RESERVED		= 0x00,
	CONNECT			= 0x01,
	CONNECTREPLY	= 0x02,
	REDIRECT		= 0x03,			/* No sample data */
	REPLY			= 0x04,
	DISCONNECT		= 0x05,
	PUSH			= 0x06,			/* No sample data */
	CONFIRMEDPUSH	= 0x07,			/* No sample data */
	SUSPEND			= 0x08,			/* No sample data */
	RESUME			= 0x09,			/* No sample data */

	GET				= 0x40,
	OPTIONS			= 0x41,			/* No sample data */
	HEAD			= 0x42,			/* No sample data */
	DELETE			= 0x43,			/* No sample data */
	TRACE			= 0x44,			/* No sample data */

	POST			= 0x60,
	PUT				= 0x61,			/* No sample data */
};

/* Warning: the encoded warnings are NOT hex but DEC */
static const value_string vals_warnings[] = {
	{ 10, "110 Resonse is stale" },
	{ 11, "111 Revalidation failed" },
	{ 12, "112 Disconnected operation" },
	{ 13, "113 Heuristic expiration" },
	{ 14, "214 Transformation applied" },
	{ 99, "199/299 Miscellaneous warning" },

	{ 0, NULL },
};

/* Modified
 * static void add_uri (proto_tree *, tvbuff_t *, guint, guint);
 */
static void add_uri (proto_tree *, tvbuff_t *);
static void add_headers (proto_tree *, tvbuff_t *);
static void add_header (proto_tree *tree, tvbuff_t *tvb, guint8 curHCP);
static guint get_value_length (tvbuff_t *, guint, guint *);
static guint add_content_type (proto_tree *, tvbuff_t *, guint, guint *);
static char * get_parameter (tvbuff_t *tvb, guint32 offset, guint32 *cnt) ;
static guint add_parameter (proto_tree *, tvbuff_t *, guint);
static void add_post_data (proto_tree *, tvbuff_t *, guint);
static void add_post_variable (proto_tree *, tvbuff_t *, guint, guint, guint, guint);
static guint add_parameter_q (proto_tree *tree, tvbuff_t *tvb, guint offset);
static void add_redirect_addresses (proto_tree *tree, tvbuff_t *tvb) ;
static void add_capabilities (proto_tree *, tvbuff_t *, char);

static void msg (char type, const char *fmt, ...);


#define MSG_INFO	0x01
#define MSG_WARN	0x02
#define MSG_ERROR	0x03
#define MSG_MODULE_NAME	__FILE__
#define MSG_MIN_LEVEL	0x02
static void msg (char type, const char *fmt, ...)
{
	va_list args;
	char	*format,
			*title;

	/* Quit if message level too low */
	if (type < MSG_MIN_LEVEL)
		return;
	
	/* Max 15 chars for title string */
	format = malloc ((strlen (MSG_MODULE_NAME) + 16 + strlen(fmt)) * sizeof (char));
	if (format)
	{
		switch (type)
		{
			case MSG_INFO:
				title = "Info";
				break;
			case MSG_WARN:
				title = "Warning";
				break;
			case MSG_ERROR:
				title = "Error";
				break;
			default:
				title = "(?)";
				break;
		}
		sprintf (format, "%s [%s]: %s\n", title, MSG_MODULE_NAME, fmt);

		va_start (args, fmt);
		vfprintf (stderr, format, args);
		va_end (args);
	}
}

/* Code to actually dissect the packets */
static void
dissect_wsp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	frame_data *fdata = pinfo->fd;
	int			offset				= 0;
	guint8		pdut;				/* Effective PDU type as encoded in the WSP PDU */
	guint8		main_pdut;			/* Main PDU type of a given WSP PDU */
	guint		tvb_len				= tvb_reported_length (tvb);
	guint		count				= 0;
	guint		value				= 0;
	guint		uriLength			= 0;
	guint		capabilityLength	= 0;
	guint		capabilityStart		= 0;
	guint		headersLength		= 0;
	guint		headerLength		= 0;
	guint		headerStart			= 0;
	guint		nextOffset			= 0;
	guint		contentTypeStart	= 0;
	guint		contentType			= 0;
	tvbuff_t	*tmp_tvb;

/* Set up structures we will need to add the protocol subtree and manage it */
	proto_item	*ti;
	proto_tree	*wsp_tree;
	
/* This field shows up as the "Info" column in the display; you should make
   it, if possible, summarize what's in the packet, so that a user looking
   at the list of packets can tell what type of packet it is. */
    
	/* Display protocol type depending on the port */
	if (check_col(fdata, COL_PROTOCOL)) 
	{
		switch ( pinfo->match_port )
		{
			case UDP_PORT_WSP:
				col_set_str(fdata, COL_PROTOCOL, "WSP" );
				break;
			case UDP_PORT_WTLS_WSP:
				col_set_str(fdata, COL_PROTOCOL, "WTLS+WSP" );
				break;
		}
	}

	/* Clear the Info column before we fetch anything from the packet */
	if (check_col(fdata, COL_INFO))
	{
		col_clear(fdata, COL_INFO);
	}

	/* Connection-less mode has a 1-octet TID first */
	if (	(pinfo->match_port == UDP_PORT_WSP) ||
			(pinfo->match_port == UDP_PORT_WTLS_WSP))
	{
		offset++;
	};

	/* Find the PDU type */
	pdut = tvb_get_guint8 (tvb, offset);

	/* Get the main PDU type */
	main_pdut = pdut;
	if ( ((0x41 <= pdut) && (pdut <= 0x44)) || ((0x50 <= pdut) && (pdut <= 0x5F)) )
		main_pdut = 0x40;
	if ( (pdut == 0x61) || ((0x70 <= pdut) && (pdut <= 0x7F)) )
		main_pdut = 0x60;

	/* Develop the string to put in the Info column */
	if (check_col(fdata, COL_INFO))
	{
		col_add_fstr(fdata, COL_INFO, "WSP %s",
			val_to_str (pdut, vals_pdu_type, "Unknown PDU type (0x%02x)"));
	};

/* In the interest of speed, if "tree" is NULL, don't do any work not
   necessary to generate protocol tree items. */
	if (tree) {
		ti = proto_tree_add_item(tree, proto_wsp,
				tvb, 0, tvb_length(tvb), bo_big_endian);
		wsp_tree = proto_item_add_subtree(ti, ett_wsp);

		/* Code to process the packet goes here */

		/* Add common items: only TID and PDU Type */

		/* If this is connectionless, then the TID Field is always first */
		if ((pinfo->match_port == UDP_PORT_WSP) ||
			(pinfo->match_port == UDP_PORT_WTLS_WSP))
		{
			ti = proto_tree_add_item (wsp_tree, hf_wsp_header_tid,
					tvb, 0, 1, bo_big_endian);
		}

		ti = proto_tree_add_item(
					wsp_tree, 					/* tree */
					hf_wsp_header_pdu_type, 	/* header field id */
					tvb, 						/* data buffer */
					offset++, 					/* start of highlighted data */
					1,							/* length of highlighted data */
					bo_big_endian				/* value encoding */
		);

		switch (main_pdut)
		{
			case CONNECT:
				/* PDU organization
				 * ================
				 * uint8	Version
				 * uintvar	CapabilitiesLen
				 * uintvar	HeadersLen
				 * *OCTET	Capabilities
				 * *OCTET	Headers
				 */
				
				/* uint8	Version */
				ti = proto_tree_add_item (wsp_tree, hf_wsp_version_major,
						tvb, offset, 1, bo_big_endian);
				ti = proto_tree_add_item (wsp_tree, hf_wsp_version_minor,
						tvb, offset, 1, bo_big_endian);
				offset++;

				/* uintvar	CapabilitiesLen */
				capabilityStart = offset;
				count = 0;	/* Initialise count */
				capabilityLength = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_capability_length,
						tvb, capabilityStart, count, capabilityLength);
				offset += count;

				/* uintvar	HeadersLen */
				headerStart = offset;
				count = 0;	/* Initialise count */
				headerLength = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_header_length,
						tvb, headerStart, count, headerLength);
				offset += count;
				
				/* *OCTET	Capabilities */
				if (capabilityLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset,
							capabilityLength, capabilityLength);
					add_capabilities (wsp_tree,tmp_tvb,CONNECT);
					offset += capabilityLength;
				}

				/* *OCTET	Headers */
				if (headerLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset,
							headerLength, headerLength);
					add_headers (wsp_tree, tmp_tvb);
				}

				break;

			case CONNECTREPLY:
				/* PDU organization
				 * ================
				 * uintvar	ServerSessionID
				 * uintvar	CapabilitiesLen
				 * uintvar	HeadersLen
				 * *OCTET	Capabilities
				 * *OCTET	Headers
				 */

				/* uintvar	ServerSessionID */
				count = 0;	/* Initialise count */
				value = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_server_session_id,
						tvb, offset, count, value);
				offset += count;

				/* uintvar	CapabilitiesLen */
				capabilityStart = offset;
				count = 0;	/* Initialise count */
				capabilityLength = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_capability_length,
						tvb, capabilityStart, count, capabilityLength);
				offset += count;

				/* uintvar	HeadersLen */
				headerStart = offset;
				count = 0;	/* Initialise count */
				headerLength = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_header_length,
						tvb, headerStart, count, headerLength);
				offset += count;

				/* *OCTET	Capabilities */
				if (capabilityLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset, capabilityLength, capabilityLength);
					add_capabilities (wsp_tree,tmp_tvb,CONNECTREPLY);
					offset += capabilityLength;
				}

				/* *OCTET	Headers */
				if (headerLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset, headerLength, headerLength);
					add_headers (wsp_tree, tmp_tvb);
				}

				break;

			case REDIRECT:
				/* PDU organization
				 * ================
				 * uint8	Flags
				 * *OCTET	Redirect address(es)
				 */

				/* uint8	Flags */
				count	= 0;
				ti = proto_tree_add_item (wsp_tree, hf_wsp_redirect_flags,
						tvb, offset, 1, bo_big_endian);
				offset++;

				/* *OCTET	Redirect address(es) */
				tmp_tvb = tvb_new_subset (tvb, offset,
						tvb_len - offset,
						tvb_len - offset);
				add_redirect_addresses (wsp_tree, tmp_tvb);
				break;
				
			case DISCONNECT:
			case SUSPEND:
				/* PDU organization
				 * ================
				 * uintvar	ServerSessionID
				 */
				
				/* uintvar	ServerSessionID */
				count = 0;	/* Initialise count */
				value = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_server_session_id,
						tvb, offset, count, value);
				break;

			case GET:
				/* PDU organization
				 * ================
				 * uintvar	URILen
				 * *OCTET	URI
				 * *OCTET	Headers
				 */
				
				/* uintvar	URILen */
				count = 0;	/* Initialise count */
				/* Length of URI and size of URILen field */
				value = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_header_uri_len,
						tvb, offset, count, value);
				offset += count;

				/* *OCTET	URI */
				if (value > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset, value, value);
					add_uri (wsp_tree, tmp_tvb);
					offset += value;
				}
				
				/* *OCTET	Headers */
				if (offset < tvb_len)
				{
					tmp_tvb = tvb_new_subset (tvb, offset, -1, -1);
					add_headers (wsp_tree, tmp_tvb);
				}
				break;

			case POST:
				/* PDU organization
				 * ================
				 * uintvar	URILen
				 * uintvar	HeadersLen
				 * *OCTET	URI
				 * *OCTET	ContentType
				 * *OCTET	Headers
				 * *OCTET	Data
				 */

				/* uintvar URILen */
				count = 0;	/* Initialise count */
				uriLength = tvb_get_guintvar (tvb, offset, &count);
				offset += count;
				
				/* uintvar HeadersLen */
				count = 0;	/* Initialise count */
				headersLength = tvb_get_guintvar (tvb, offset, &count);
				offset += count;

				/* *OCTET	URI */
				add_uri (wsp_tree, tvb);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_header_uri_len,
						tvb, offset, count, value);
				offset += uriLength;

				ti = proto_tree_add_item (wsp_tree, hf_wsp_header_length,
						tvb, headerStart, count, bo_big_endian);

				/* *OCTET	ContentType */
				contentTypeStart = offset;
				nextOffset = add_content_type (wsp_tree, tvb, offset, &contentType);

				/* *OCTET	Headers */
				/* Add headers subtree that will hold the headers fields */
				/* Runs from nextOffset for value-(length of content-type field)*/
				headerLength = headersLength - (nextOffset - contentTypeStart);
				if (headerLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, nextOffset, headerLength, headerLength);
					add_headers (wsp_tree, tmp_tvb);
					offset = nextOffset + headerLength;
				}

				/* *OCTET	Data */
				/* TODO: Post DATA */
				/* Runs from start of headers+headerLength to end of frame */
				if (offset < tvb_len)
				{
					tmp_tvb = tvb_new_subset (tvb, offset,
							tvb_len - offset, tvb_len - offset);
					add_post_data (wsp_tree, tmp_tvb, contentType);
				}
				break;

			case REPLY:
				/* PDU organization
				 * ================
				 * uint8	Status
				 * uintvar	HeadersLen
				 * *OCTET	ContentType
				 * *OCTET	Headers
				 * *OCTET	Data
				 */

				/* uint8	Status */
				ti = proto_tree_add_item (wsp_tree, hf_wsp_header_status,
						tvb, offset, 1, bo_big_endian);
				offset++;

				/* uintvar	HeadersLen */
				count = 0;	/* Initialise count */
				value = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_header_length,
						tvb, offset, count, value);
				offset += count;

				/* *OCTET	ContentType */
				contentTypeStart = offset;
				if (value > 0)
				{
					offset = add_content_type (wsp_tree, tvb, offset, &contentType);
				}

				/* *OCTET	Headers */
				/* Add headers subtree that will hold the headers fields */
				/* Runs from nextOffset for value - (length of content - type field) */
				headerLength = value - (offset - contentTypeStart);
				if (headerLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset, headerLength, headerLength);
					add_headers (wsp_tree, tmp_tvb);
				}
				offset = contentTypeStart + value;

				/* *OCTET	Data */
				/* TODO: Data - decode WMLC */
				/* Runs from offset+1+count+value+1 to end of frame */
				if (offset < tvb_len)
				{
					ti = proto_tree_add_item (wsp_tree, hf_wsp_reply_data,
							tvb, offset, tvb_length_remaining(tvb, offset), bo_big_endian);
				}
				break;

			case PUSH:
			case CONFIRMEDPUSH:
				/* PDU organization
				 * ================
				 * uintvar	HeadersLen
				 * *OCTET	ContentType
				 * *OCTET	Headers
				 * *OCTET	Data
				 */
				
				/* uintvar	HeadersLen */
				count = 0;	/* Initialise count */
				value = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_header_length,
						tvb, offset, count, value);
				offset += count;

				/* *OCTET	ContentType */
				contentTypeStart = offset;
				if (value > 0)
				{
					offset = add_content_type (wsp_tree, tvb, offset, &contentType);
				}

				/* *OCTET	Headers */
				/* Add headers subtree that will hold the headers fields */
				/* Runs from nextOffset for value - (length of content - type field) */
				headerLength = value - (offset - contentTypeStart);
				if (headerLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset, headerLength, headerLength);
					add_headers (wsp_tree, tmp_tvb);
				}
				offset = contentTypeStart + value;

				/* *OCTET	Data */
				/* TODO: Data - decode WMLC */
				/* Runs from offset+1+count+value+1 to end of frame */
				if (offset < tvb_len)
				{
					ti = proto_tree_add_item (wsp_tree, hf_wsp_reply_data,
							tvb, offset, tvb_length_remaining(tvb, offset), bo_big_endian);
				}
				break;

			case RESUME:
				/* PDU organization
				 * ================
				 * uintvar	SessionID
				 * uintvar	CapabilitiesLen
				 * *OCTET	Capabilities
				 * *OCTET	Headers
				 */

				/* uintvar	SessionID */
				count = 0;	/* Initialise count */
				value = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_server_session_id,
						tvb, offset, count, value);
				offset += count;

				/* uintvar	CapabilitiesLen */
				capabilityStart = offset;
				count = 0;	/* Initialise count */
				capabilityLength = tvb_get_guintvar (tvb, offset, &count);
				ti = proto_tree_add_uint (wsp_tree, hf_wsp_capability_length,
						tvb, capabilityStart, count, capabilityLength);
				offset += count;

				/* *OCTET	Capabilities */
				if (capabilityLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset, capabilityLength, capabilityLength);
					add_capabilities (wsp_tree,tmp_tvb,CONNECTREPLY);
					offset += capabilityLength;
				}

				/* *OCTET	Headers */
				headerLength = tvb_len - offset;
				if (headerLength > 0)
				{
					tmp_tvb = tvb_new_subset (tvb, offset, headerLength, headerLength);
					add_headers (wsp_tree, tmp_tvb);
				}
				break;
		}
	}
}



/* Adds an URI with a given length */
static void
add_uri (proto_tree *tree, tvbuff_t *tvb)
{
	proto_item	*ti;
	guint		uriLen	= tvb_reported_length (tvb);

	/* If string doesn't end with a 0x00, we need to add one to be on the safe side */
	if (tvb_get_guint8 (tvb, uriLen - 1) != 0)
	{
		char *newBuffer = g_malloc (uriLen + 1);
		strncpy (newBuffer, tvb_get_ptr (tvb, 0, uriLen), uriLen);
		newBuffer[uriLen] = 0;
		ti = proto_tree_add_string (tree, hf_wsp_header_uri, tvb, 0, uriLen, newBuffer);
		g_free (newBuffer);
	}
	else
	{
		ti = proto_tree_add_item (tree, hf_wsp_header_uri, tvb, 0, uriLen, bo_big_endian);
	}
}





static void
add_headers (proto_tree *tree, tvbuff_t *tvb)
{
	proto_item	*ti;
	proto_tree	*wsp_headers;
	guint		len				= 0,
				offset			= 0;
	guint8		curHCP			= 1; /* Standard Header Code Page */
	tvbuff_t	*hdr_buf		= NULL;

	if (! tree)
	{
		msg (MSG_INFO, "add_headers: tree == NULL");
		return;
	}
	if (! tvb)
	{
		msg (MSG_INFO, "add_headers: tvb == NULL");
		return;
	}
	
	len = tvb_reported_length (tvb);
	if (len <= 0) {
		return;
	}

	ti			= proto_tree_add_item (tree, hf_wsp_headers_section, tvb, offset, len, bo_big_endian);
	wsp_headers = proto_item_add_subtree (ti, ett_headers);

	/* Start parsing the headers tvb for header entries */
	while (offset < len) {
		guint	start	= offset;
		guint8	peek	= tvb_get_guint8 (tvb, offset);

		/* Header Decoding */
		if (peek <= 31) { /* Short-cut shift to the given Header Code Page */
			curHCP = peek;
			proto_tree_add_text (wsp_headers, tvb, offset, 1,
					"Shift to Header Code Page 0x%02X", curHCP);
			offset++;
		} else if (peek == 127) { /* Shift to the given Header Code Page */
			curHCP	= tvb_get_guint8 (tvb, offset + 1);
			proto_tree_add_text (wsp_headers, tvb, offset, 2,
					"Shift to Header Code Page 0x%02X", curHCP);
			offset	+= 2;
		} else { /* Header */
			if (peek <= 126) { /* Textual header encoding */
				while ( tvb_get_guint8 (tvb, ++offset) )
					; /* Skip till end of string */
				offset++; /* Skip \0 */
			} else { /* Well-known header */
				offset++;
			}

			/* Header Value Decoding */
			peek = tvb_get_guint8 (tvb, offset);
			if (peek <= 30) { /* value length */
				offset += 1 + peek;
			} else if (peek == 31) { /* value length in following uintvar */
				guint	count		= 0;
				guint32 valueLen	= tvb_get_guintvar (tvb, offset+1, &count);
				offset += 1 + count + valueLen;
			} else if ((32 <= peek) && (peek <= 127)) { /* Textual value encoding */
				while ( tvb_get_guint8 (tvb, ++offset) )
					; /* Skip till end of string */
				offset++; /* Skip \0 */
			} else { /* Well-known value */
				offset++;
			}

			/* Add the given header */
			hdr_buf = tvb_new_subset (tvb, start, offset - start, offset - start);
			add_header (wsp_headers, hdr_buf, curHCP);
		}
	}
}




#include "packet-wsp-bnf.c"
#include "packet-wsp-headers.h"

/* The following array defines the unique IDs of well-known-header values.
 * Since [WSP] redefined some header IDs in [WSP 1.3], there was no unique
 * mapping anymore.  This is again obtained with the array below.
 */
static const guint8 unique_wsp_header_id[] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
		0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
		0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
		0x38, 0x39, 0x3A, 0x01, 0x02, 0x08, 0x10, 0x3F,
	0x40, 0x41, 0x42, 0x43, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
};


static void
add_header (proto_tree *tree, tvbuff_t *tvb, guint8 curHCP)
{
	proto_item	*ti;
	guint		len				= 0,
				offset			= 0;
	guint8		hdrType			= 0xFF;
	guint8		hdrID			= 0xFF;
	char		*hdrString		= NULL;
	guint32		valType			= 0;
	guint32		valStart		= 0;
	guint32		valLen			= 0;
	char		*valString		= NULL;

	msg (MSG_INFO, "add_header : START");
	if (! tree) {
		msg (MSG_INFO, "add_header: tree == NULL");
		return;
	}
	if (! tvb) {
		msg (MSG_INFO, "add_header: tvb == NULL");
		return;
	}
	
	len = tvb_reported_length (tvb);
	if (len <= 0) {
		return;
	}

	/* Start parsing the headers tvb for header entries */
	hdrType		= tvb_get_guint8 (tvb, offset);
	msg (MSG_INFO, "add_header: found header 1st byte 0x%02X, length = %d",
			hdrType, len);

	/* Header Decoding */
	if (hdrType <= 126) { /* Textual header encoding */
		hdrString	= (char *)tvb_get_ptr (tvb, 0, len);
		valStart	= strlen (hdrString) + 1;
	} else { /* Well-known header */
		hdrID		= unique_wsp_header_id[hdrType & 0x7F];
		hdrString	= match_strval (hdrID, vals_header_names);
		valStart	= 1;
	}
	msg (MSG_INFO, "add_header: Type = 0x%02X [%s], value start %d",
			hdrType, hdrString, valStart);
	if (hdrString) {
		char	*s	= NULL;
		guint	len = strlen (hdrString);
		s = malloc ((len + 1) * sizeof (char));
		if (s) {
			guint i;
			for (i = 0; i < len; i++) {
				s[i] = toupper (hdrString[i]);
			}
		}
		hdrID	= get_wsp_hdr_id (s, len);
	}
	offset = valStart;
	/* Now the following entities are defined:
	 * 
	 * hdrType		: original well-known header (1 OCTET)
	 * hdrString	: original textual header (NULL terminated string)
	 * hdr			: {value, strptr} if header is known by [WSP]
	 *
	 * offset		: start of value chunk
	 */

	/* Header Value Decoding */
	valType = tvb_get_guint8 (tvb, offset);
	msg (MSG_INFO, "add_header: found value 1st byte 0x%02X",
			valType);
	if (valType <= 30) { /* value length */
		valStart	= 1 + offset;
		valLen		= valType;
	} else if (valType == 31) { /* value length in following uintvar */
		guint	count	= 0;
		valLen	= tvb_get_guintvar (tvb, offset+1, &count);
		valStart	= 1 + offset + count;
	} else if ((32 <= valType) && (valType <= 127)) { /* Textual value encoding */
		valString	= (char *)tvb_get_ptr (tvb, valStart, len - valStart);
		valLen		= strlen (valString) + 1;
	} else { /* Well-known value */
		valLen		= 1;
	}
	msg (MSG_INFO, "add_header: valType = 0x%02X , value start %d, value length %d",
			valType, valStart, valLen);
	/* Now the following entities are defined:
	 *
	 * valType		: 1st OCTET
	 * valStart		: start offset for value data
	 * valLen		: effective net length of value data
	 * valString	: textual value (NULL terminated string)
	 */

	/* Add the given header */

	
	switch (curHCP) {
		
		case 1:	/* Default HCP */
			switch (hdrID) {
				
				/* Accept */
				case WSP_HDR_ACCEPT:
					if (valType & 0x80) {
						ti = proto_tree_add_uint (tree, hf_wsp_header_accept,
								tvb, 0, len, valType & 0x7F);
					} else if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_accept_str,
								tvb, 0, len, valString);
					} else { /* Accept-general-form */
						guint	off		= valStart;
						guint8	peek	= tvb_get_guint8 (tvb, off);
						
						/* Short-int OR Long-int OR Text-string */
						if (peek & 0x80) { /* Short-integer */
							ti = proto_tree_add_uint (tree, hf_wsp_header_accept,
									tvb, 0, len, peek & 0x7F);
							off++;
						} else if (peek < 0x1F) { /* Long-integer */
							guint	count	= 0;
							guint32	*s		= get_bnf_long_integer (tvb, off, &count);
							
							if (s) {
								if (count <= 4) { /* 24 bit max */
									ti = proto_tree_add_uint (tree, hf_wsp_header_accept,
											tvb, 0, len, *s);
								} else { /* Error: too long for 24 bit */
									msg (MSG_WARN, "Accept value not decoded");
								}
								free (s);
								off += count;
							} else { /* Error: s */
								off = len;	/* Skip further processing */
							}
						}

						if (off < len) { /* Check for params */
							while (off < len)
							{
								/* add_parameter */
								off = add_parameter (tree, tvb, off);
							}
						}
					}	
					break;

				/* Accept-Charset */
				case WSP_HDR_ACCEPT_CHARSET:
					if (valType & 0x80) {
						ti = proto_tree_add_uint (tree, hf_wsp_header_accept_charset,
								tvb, 0, len, valType & 0x7F);
					} else if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_accept_charset_str,
								tvb, 0, len, valString);
					} else { /* Accept-charset-general-form */
						guint	off		= valStart;
						guint8	peek	= tvb_get_guint8 (tvb, off);
						
						/* Short-int OR Long-int OR Token-text */
						if (peek & 0x80) {
							ti = proto_tree_add_uint (tree, hf_wsp_header_accept_charset,
									tvb, 0, len, peek & 0x7F);
							off++;
						} else if (peek < 0x1F) { /* Long-integer */
							guint	count	= 0;
							guint32	*s		= get_bnf_long_integer (tvb, off, &count);
							if (s) {
								if (count <= 4) { /* 24 bit max */
									ti = proto_tree_add_uint (tree, hf_wsp_header_accept_charset,
											tvb, 0, len, (*s));
								} else { /* Too long for 24-bit int */
									msg (MSG_WARN, "Accept-Charset value not decoded");
								}
								free (s);
								off += count;
							} else { /* Error: s */
								off = len;	/* Skip further processing */
							}
						}
						if (off < len) { /* Check for params */
							if (off < len)
							{
								/* add_parameter */
								off = add_parameter_q (tree, tvb, off);
							}
						}
					}	
					break;

				/* Accept-Encoding */
				case WSP_HDR_ACCEPT_ENCODING:
					if (valType & 0x80) {
						ti = proto_tree_add_uint (tree, hf_wsp_header_accept_encoding,
								tvb, 0, len, valType & 0x7F);
					} else if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_accept_encoding_str,
								tvb, 0, len, valString);
					} else { /* Accept-encoding-general-form */
						guint	off		= valStart;
						guint8	peek	= tvb_get_guint8 (tvb, off);
						
						/* Short-int OR Token-text */
						if (peek & 0x80) {
							ti = proto_tree_add_uint (tree, hf_wsp_header_accept_encoding,
									tvb, 0, len, peek & 0x7F);
							off++;
						} else { /* Token-text */
							guint	count	= 0;
							char	*s		= get_bnf_text_string (tvb, off, &count);
							if (s) {
								ti = proto_tree_add_string (tree, hf_wsp_header_accept_encoding_str,
										tvb, 0, len, s);
								free (s);
								off += count;
							} else { /* Error: s */
								off = len;	/* Skip further processing */
							}
						}
						if (off < len) { /* Check for params */
							if (off < len)
							{
								/* add_parameter */
								off = add_parameter_q (tree, tvb, off);
							}
						}
					}	
					break;
					
				/* Content-Encoding */
				case WSP_HDR_CONTENT_ENCODING:
					if (valType & 0x80) { /* Short-int */
						ti = proto_tree_add_uint (tree, hf_wsp_header_content_encoding,
								tvb, 0, len, valType & 0x7F);
					} else if (istokenchar (valType)) { /* Token-text */
						if (valString) {
							ti = proto_tree_add_string (tree, hf_wsp_header_content_encoding_str,
									tvb, 0, len, valString);
						} else { /* Error */ }
					} else { /* Error: invalid encoding */
					}	
					break;


				/* Accept-Language */
				case WSP_HDR_ACCEPT_LANGUAGE:
					if (valType & 0x80) {
						ti = proto_tree_add_uint (tree, hf_wsp_header_accept_language,
								tvb, 0, len, valType & 0x7F);
					} else if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_accept_language_str,
								tvb, 0, len, valString);
					} else { /* Accept-language-general-form */
						guint	off		= valStart;
						guint8	peek	= tvb_get_guint8 (tvb, off);
						
						/* Short-int OR Long-int OR Token-text */
						if (peek & 0x80) {
							ti = proto_tree_add_uint (tree, hf_wsp_header_accept_language,
									tvb, 0, len, peek & 0x7F);
							off++;
						} else if (peek < 0x1F) { /* Long-integer */
							guint	count	= 0;
							guint32	*s		= get_bnf_long_integer (tvb, off, &count);
							if (s) {
								if (count <= 5) { /* 1 + max 4 */
									ti = proto_tree_add_uint (tree, hf_wsp_header_accept_language,
											tvb, 0, len, (*s));
								} else { /* Too long for 32-bit int */
									msg (MSG_WARN, "Accept-Language value not decoded");
								}
								free (s);
								off += count;
							} else { /* Error: s */
								off = len;	/* Skip further processing */
							}
						}
						if (off < len) { /* Check for params */
							if (off < len)
							{
								/* add_parameter */
								off = add_parameter_q (tree, tvb, off);
							}
						}
					}	
					break;

				/* Content-Language */
				case WSP_HDR_CONTENT_LANGUAGE:
					if (valType & 0x80) { /* Short-int */
						ti = proto_tree_add_uint (tree, hf_wsp_header_content_language,
								tvb, 0, len, valType & 0x7F);
					} else if (valType < 0x1F) { /* Long-int */
						guint32	count	= 0;
						guint32	*s		= get_bnf_long_integer (tvb, valStart, &count);

						if (s) {
							if (count <= 5) { /* Max 16 bits decoded */
								ti = proto_tree_add_uint (tree, hf_wsp_header_content_language,
										tvb, 0, len, (*s));
							} else { /* Error: too big number */
								msg (MSG_WARN, "Content-Language value not decoded");
							}
							free (s);
						} else { /* Error */ }
					} else if (istokenchar (valType)) {
						if (valString) {
							ti = proto_tree_add_string (tree, hf_wsp_header_content_language_str,
									tvb, 0, len, valString);
						} else { /* Error */ }
					} else { /* Error: wrong encoding */
					}
					break;


				/* text string */
				case WSP_HDR_ETAG:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_etag,
								tvb, 0, len, valString);
					}
					break;

				case WSP_HDR_FROM:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_from,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_HOST:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_host,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_IF_MATCH:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_if_match,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_IF_NONE_MATCH:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_if_none_match,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_SERVER:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_server,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_UPGRADE:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_upgrade,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_USER_AGENT:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_user_agent,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_VIA:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_via,
								tvb, 0, len, valString);
					}
					break;

				/* quoted text string */
				case WSP_HDR_CONTENT_ID:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_content_id,
								tvb, 0, len, valString);
					}
					break;

				/* URI (text string) */
				case WSP_HDR_CONTENT_BASE:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_content_base,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_CONTENT_LOCATION:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_content_location,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_LOCATION:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_location,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_REFERER:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_referer,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_X_WAP_CONTENT_URI:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_x_wap_content_uri,
								tvb, 0, len, valString);
					}
					break;
					
				case WSP_HDR_X_WAP_INITIATOR_URI:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_x_wap_initiator_uri,
								tvb, 0, len, valString);
					}
					break;

				case WSP_HDR_PROFILE:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_profile,
								tvb, 0, len, valString);
					}
					break;



				/* Date-value */
				case WSP_HDR_DATE:
					if (valType <= 30) {
						guint count = 0;
						struct timeval tv;
						guint32 *s = get_bnf_long_integer (tvb, offset, &count);

						if (s) {
							if (count <= 5) { /* 1 + max 4 */
								tv.tv_sec = (*s);
								tv.tv_usec = 0;
								ti = proto_tree_add_time (tree, hf_wsp_header_date,
										tvb, 0, len, &tv);
							} else { /* Error: too big number */
								msg (MSG_WARN, "Date: Date value not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					}
					break;

				case WSP_HDR_EXPIRES:
					if (valType <= 30) {
						guint count = 0;
						struct timeval tv;
						guint32 *s = get_bnf_long_integer (tvb, offset, &count);

						if (s) {
							if (count <= 5) { /* 1 + max 4 */
								tv.tv_sec = (*s);
								tv.tv_usec = 0;
								ti = proto_tree_add_time (tree, hf_wsp_header_expires,
										tvb, 0, len, &tv);
							} else { /* Error: too big number */
								msg (MSG_WARN, "Expires: Date value not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					}
					break;

				case WSP_HDR_IF_MODIFIED_SINCE:
					if (valType <= 30) {
						guint count = 0;
						struct timeval tv;
						guint32 *s = get_bnf_long_integer (tvb, offset, &count);

						if (s) {
							if (count <= 5) { /* 1 + max 4 */
								tv.tv_sec = (*s);
								tv.tv_usec = 0;
								ti = proto_tree_add_time (tree, hf_wsp_header_if_modified_since,
										tvb, 0, len, &tv);
							} else { /* Error: too big number */
								msg (MSG_WARN, "If-Modified-Since: Date value not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					}
					break;

				case WSP_HDR_IF_UNMODIFIED_SINCE:
					if (valType <= 30) {
						guint count = 0;
						struct timeval tv;
						guint32 *s = get_bnf_long_integer (tvb, offset, &count);

						if (s) {
							if (count <= 5) { /* 1 + max 4 */
								tv.tv_sec = (*s);
								tv.tv_usec = 0;
								ti = proto_tree_add_time (tree, hf_wsp_header_if_unmodified_since,
										tvb, 0, len, &tv);
							} else { /* Error: too big number */
								msg (MSG_WARN, "If-Unmodified-Since: Date value not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					}
					break;

				case WSP_HDR_LAST_MODIFIED:
					if (valType <= 30) {
						guint count = 0;
						struct timeval tv;
						guint32 *s = get_bnf_long_integer (tvb, offset, &count);

						if (s) {
							if (count <= 5) { /* 1 + max 4 */
								tv.tv_sec = (*s);
								tv.tv_usec = 0;
								ti = proto_tree_add_time (tree, hf_wsp_header_last_modified,
										tvb, 0, len, &tv);
							} else { /* Error: too big number */
								msg (MSG_WARN, "Last-Modified: Date value not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					}
					break;

				case WSP_HDR_X_WAP_TOD:
					if (valType <= 30) {
						guint count = 0;
						struct timeval tv;
						guint32 *s = get_bnf_long_integer (tvb, offset, &count);

						if (s) {
							if (count <= 5) { /* 1 + max 4 */
								tv.tv_sec = (*s);
								tv.tv_usec = 0;
								ti = proto_tree_add_time (tree, hf_wsp_header_x_wap_tod,
										tvb, 0, len, &tv);
							} else { /* Error: too big number */
								msg (MSG_WARN, "X-Wap-Tod: Date value not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					}
					break;

				/* Date-value or Text-string */
				case WSP_HDR_IF_RANGE:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_if_range_str,
								tvb, 0, len, valString);
					} else if (valType <= 30) {
						guint count = 0;
						struct timeval tv;
						guint32 *s = get_bnf_long_integer (tvb, offset, &count);

						if (s) {
							if (count <= 5) { /* 1 + max 4 */
								tv.tv_sec = (*s);
								tv.tv_usec = 0;
								ti = proto_tree_add_time (tree, hf_wsp_header_if_range_date,
										tvb, 0, len, &tv);
							} else { /* Error: too big number */
								msg (MSG_WARN, "If-Range: Date value not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					}
					break;

				case WSP_HDR_ACCEPT_RANGES:
					if (valString) {
						ti = proto_tree_add_string (tree, hf_wsp_header_accept_ranges_str,
								tvb, 0, len, valString);
					} else if (valType & 0x80) {
						ti = proto_tree_add_uint (tree, hf_wsp_header_accept_ranges,
								tvb, 0, len, valType);
					}
					break;

				case WSP_HDR_WARNING:
					msg (MSG_INFO, "Warning header: START in SWITCH()");
					if (valType & 0x80) { /* Warn-code */
						msg (MSG_INFO, "Warning header: short (warn_code)");
						ti = proto_tree_add_uint (tree, hf_wsp_header_warning,
								tvb, 0, len, hdrType & 0x7F);
					} else { /* Warning-general-form */
						guint	off		= valStart;
						guint	count	= 0;
						guint8	peek	= tvb_get_guint8 (tvb, off);
						char	*warn	= match_strval (peek & 0x7F, vals_warnings);
						char	*warn_agent;
						char	*warn_text;
						
						msg (MSG_INFO, "Warning header: general form");
						/* Short-int Text-string Text-string */
						off++;
						msg (MSG_INFO, "Warning header: general form: warn_agent");
						if ( (warn_agent = get_bnf_text_string (tvb, off, &count)) ) {
							off += count;
							if (off < len) {
								count	= 0;
								msg (MSG_INFO, "Warning header: general form: warn_text");
								if ( (warn_text = get_bnf_text_string (tvb, off, &count)) ) {
									char *s;
									off += count;

									s = malloc (((off - valStart) + strlen (warn) + 30) * sizeof (char));
									if (s) {
										sprintf (s, "%s (Agent='%s', Text='%s')", warn, warn_agent, warn_text);
										ti = proto_tree_add_string (tree, hf_wsp_header_warning_str,
												tvb, 0, len, s);
										free (s);
									} else {
										msg (MSG_ERROR, "add_header: Long-warn-code: malloc() failed.");
									}
									free (warn_text);
								}
								free (warn_agent);
							}
						}

					}
					break;

				case WSP_HDR_AGE:
					if (valType & 0x80) {
						struct timeval tv;
						
						tv.tv_sec = valType & 0x7F;
						tv.tv_usec = 0;
						ti = proto_tree_add_time (tree, hf_wsp_header_age,
								tvb, 0, len, &tv);
					} else if (valType <= 30) { /* Integer-value */
						guint count = 0;
						struct timeval tv;
						guint32 *s = get_bnf_long_integer (tvb, offset, &count);

						if (s) {
							if (count <= 5) { /* 1 + max 4 */
								tv.tv_sec = (*s);
								tv.tv_usec = 0;
								ti = proto_tree_add_time (tree, hf_wsp_header_age,
										tvb, 0, len, &tv);
							} else { /* Error: too big number */
								msg (MSG_WARN, "Age: Date value not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					}
					break;

				case WSP_HDR_ALLOW:
					/* Short-integer: method */
					if (valType & 0x80) {
						if (valType >= 0xC0) { /* Valid method */
							ti = proto_tree_add_uint (tree, hf_wsp_header_allow,
									tvb, 0, len, valType & 0x7F);
						} else { /* Invalid method! */
							ti = proto_tree_add_uint (tree, hf_wsp_header_allow,
									tvb, 0, len, 0);
						}
					} else {
						msg (MSG_WARN, "add_header: Allow header incorrectly encoded in packet!");
					}
					break;

				case WSP_HDR_EXPECT:
					if (valType & 0x80) { /* 100-continue */
						if (valType == 0x80) {
							ti = proto_tree_add_string (tree, hf_wsp_header_expect_text,
									tvb, 0, len, "100 Continue");
						} else { /* Error: wrong code */
							msg (MSG_WARN, "add_header: Expect header: invalid type 0x%02X !", valType);
						}
					} else { /* Value-length Token-text (Token-text|Quoted-string) *(Token-text (Token-text|Quoted-string)) */
						guint32 off = valStart;
						char	*var	= NULL;
						char	*val	= NULL;
						guint32	varLen	= 0;
						guint32	valLen	= 0;
						guint32 count	= 0;

						/* Token-text (Token-text | Quoted-string) */
						var = get_bnf_token_text (tvb, off, &count);
						varLen = count;
						off += count;
						if (var) {
							if (off < len) {
								count = 0;
								val = get_bnf_text_string (tvb, off, &count);
								valLen = count;
								off += count;
								
								if (val) {
									char *s;
									s = malloc ((varLen + 1 + valLen + 2) * sizeof (char));
									if (s) {
										sprintf (s, "%s %s", var, val);
										ti = proto_tree_add_string (tree, hf_wsp_header_expect_text,
												tvb, 0, len, s);
										free (s);
										/* *(Token-text (Token-text | Quoted-string)) */
										while (off < len) { /* Expect parameter(s) */
											count = 0;
											var = get_bnf_token_text (tvb, off, &count);
											varLen = count;
											off += count;
											if (var) {
												if (off < len) {
													count = 0;
													val = get_bnf_text_string (tvb, off, &count);
													valLen = count;
													off += count;
													if (val) {
														s = malloc ((varLen + 1 + valLen + 2) * sizeof (char));
														if (s) {
															sprintf (s, "%s %s", var, val);
															ti = proto_tree_add_string (tree, hf_wsp_header_expect_parm,
																	tvb, 0, len, s);
															free (s);
														} else { /* Error */ }
													} else { /* Error */ }
												} else { /* Error */ }
											} else { /* Error */ }
										} /* while */
									} else { /* malloc failed */ }
								} else { /* Error */ }
							} else { /* Buffer overrun */ }
						} else { /* Error */ }
					}
					break;

				case WSP_HDR_PROFILE_WARNING:
					if (valType & 0x80) { /* Short-int */
						ti = proto_tree_add_uint (tree, hf_wsp_header_profile_warning_code,
								tvb, 0, len, valType & 0x7F);
					} else { /* Value-length Token-text (Token-text|Quoted-string) *(Token-text (Token-text|Quoted-string)) */
						guint32 off			= valStart;
						char	*target		= NULL;
						guint32 count		= 0;
						guint8	code		= tvb_get_guint8 (tvb, off);

						msg (MSG_WARN, "add_header: Profile-Warning header: long format not yet parsed!");

						/* Short-int Uri-value *Date-value */
						ti = proto_tree_add_uint (tree, hf_wsp_header_profile_warning_code,
								tvb, 0, len, code & 0x7F);
						off++;
						target = get_bnf_text_string (tvb, off, &count);
						if (target) {
							ti = proto_tree_add_string (tree, hf_wsp_header_profile_warning_uri,
									tvb, off, count, target);
							off += count;
							while (off < len) { /* *Date-value */
								guint32 count = 0;
								struct timeval tv;
								guint32 *s = get_bnf_long_integer (tvb, off, &count);

								if (count <= 5) { /* 1 + max 4 */
									tv.tv_sec = (*s);
									tv.tv_usec = 0;
									ti = proto_tree_add_time (tree, hf_wsp_header_profile_warning_date,
											tvb, off, count, &tv);
								} else { /* Error: too big number */
									msg (MSG_WARN, "Profile-Warning: Date value not decoded");
								}
								off += count;
							}
						} else { /* Error */
						}
					}
					break;

				case WSP_HDR_CACHE_CONTROL:
					/* Short-int | Token-text */
					/* Value-length (Short-int | Token-text) (specific values) */
					if (valType & 0x80) { /* Short-int */
						ti = proto_tree_add_uint (tree, hf_wsp_header_cache_control,
								tvb, 0, len, valType);
					} else if (istokenchar(valType)) { /* Token-text */
						if (valString) {
							ti = proto_tree_add_string (tree, hf_wsp_header_cache_control_tok,
										tvb, 0, len, valString);
						} else { /* Error */ }
					} else if (valType <= 0x1F) { /* (Short-int | Token-text) (specific values) */
						guint32	off		= valStart;
						guint8	code	= tvb_get_guint8 (tvb, off);
						msg (MSG_INFO, "add_header: Cache-Control header: code = %02X, valStart = %d, len = %d", code, off, len);
						off++;
						if (code & 0x80) {
							switch (code) {
								case 0x82: /* Max-Age */
									if (off < len) { /* Delta-seconds-value */
										guint32 count = 0;
										struct timeval tv;
										guint32 *s = get_bnf_integer_value (tvb, off, &count);
		
										if (s) {
											if (count <= 5) { /* 1 + max 4 */
												tv.tv_sec = (*s);
												tv.tv_usec = 0;
												ti = proto_tree_add_time (tree, hf_wsp_header_cache_control_max_age,
														tvb, 0, len, &tv);
											} else { /* Too large time value */
												msg (MSG_WARN, "Cache:Control: Max-Age: too large time value (%d chars)", count);
											}
											off += count;
											free (s);
										} else { /* Error: s */ }
									} else { /* Error */ }
									break;
								case 0x83: /* Max-Stale */
									if (off < len) { /* Delta-seconds-value */
										guint32 count = 0;
										struct timeval tv;
										guint32 *s = get_bnf_integer_value (tvb, off, &count);
		
										if (s) {
											if (count <= 5) { /* 1 + max 4 */
												tv.tv_sec = (*s);
												tv.tv_usec = 0;
												ti = proto_tree_add_time (tree, hf_wsp_header_cache_control_max_stale,
														tvb, 0, len, &tv);
											} else { /* Too large time value */
												msg (MSG_WARN, "Cache:Control: Max-Stale: too large time value (%d chars)", count);
											}
											off += count;
											free (s);
										} else { /* Error: s */ }
									} else { /* Error */ }
									break;
								case 0x84: /* Min-Fresh */
									if (off < len) { /* Delta-seconds-value */
										guint32 count = 0;
										struct timeval tv;
										guint32 *s = get_bnf_integer_value (tvb, off, &count);
		
										if (s) {
											if (count <= 5) { /* 1 + max 4 */
												tv.tv_sec = (*s);
												tv.tv_usec = 0;
												ti = proto_tree_add_time (tree, hf_wsp_header_cache_control_min_fresh,
														tvb, 0, len, &tv);
											} else { /* Too large time value */
												msg (MSG_WARN, "Cache:Control: Min-Fresh: too large time value (%d chars)", count);
											}
											off += count;
											free (s);
										} else { /* Error: s */ }
									} else { /* Error */ }
									break;
									
								case 0x80: /* No-Cache */
									while (off < len) { /* 1 or more field names */
										/* Field-name := Well-known-field-name | Token-text
										 * Well-known-field-name := Short-integer
										 */
										guint8 peek = tvb_get_guint8 (tvb, off);
										if (peek & 0x80) { /* Well-known-header */
											ti = proto_tree_add_uint (tree, hf_wsp_header_cache_control_no_cache,
													tvb, 0, len, peek & 0x7F);
										} else if (istokenchar (peek)) {
											guint32	count	= 0;
											char	*s		= get_bnf_text_string (tvb, off, &count);
											if (s) {
												ti = proto_tree_add_string (tree, hf_wsp_header_cache_control_no_cache_str,
														tvb, 0, len, s);
											} else { /* Error: s */ }
										} else { /* Error: invalid coding */
											msg (MSG_WARN, "add_header: Cache-Control: No-Cache: Invalid coding (Type = %02X)", peek);
										}
									}
									break;
								case 0x87: /* Private */
									while (off < len) { /* 1 or more field names */
										/* Field-name := Well-known-field-name | Token-text
										 * Well-known-field-name := Short-integer
										 */
										guint8 peek = tvb_get_guint8 (tvb, off);
										if (peek & 0x80) { /* Well-known-header */
											ti = proto_tree_add_uint (tree, hf_wsp_header_cache_control_private,
													tvb, 0, len, peek & 0x7F);
										} else if (istokenchar (peek)) {
											guint32	count	= 0;
											char	*s		= get_bnf_text_string (tvb, off, &count);
											if (s) {
												ti = proto_tree_add_string (tree, hf_wsp_header_cache_control_private_str,
														tvb, 0, len, s);
											} else { /* Error: s */ }
										} else { /* Error: invalid coding */
											msg (MSG_WARN, "add_header: Cache-Control: Private: Invalid coding (Type = %02X)", peek);
										}
									}
									break;
									
								default:
									msg (MSG_WARN, "add_header: Cache-Control: Invalid coding (code = %02X)", code);
									break;
							}
						} else if (istokenchar (code)) {
							guint32	count	= 0;
							char	*s		= get_bnf_text_string (tvb, off, &count);

							if (s) {
								ti = proto_tree_add_string (tree, hf_wsp_header_cache_control_tok,
										tvb, 0, len, s);
								off += count;
								while (off < len) {
									off = add_parameter (tree, tvb, off);
								}
							} else { /* Error: s */ }
						} else { /* Error */
							msg (MSG_WARN, "add_header: Cache-Control: Invalid coding (valType = %02X)", valType);
						}
					}
					break;

				/* Integer-value */
				case WSP_HDR_CONTENT_LENGTH:
					if ((valType & 0x80) || (valType < 0x1F)) {
						guint32	count	= 0;
						guint32	*s		= get_bnf_integer_value (tvb, valStart, &count);
						if (s) {
							ti = proto_tree_add_uint (tree, hf_wsp_header_content_length,
									tvb, 0, len, (*s));
							free (s);
						} else { /* Error: s */ }
					} else { /* Error */
						msg (MSG_WARN, "add_header: Content-Length: Invalid encoding (valType = %02X", valType);
					}
					break;

				case WSP_HDR_MAX_FORWARDS:
					if ((valType & 0x80) || (valType < 0x1F)) {
						guint32	count	= 0;
						guint32	*s		= get_bnf_integer_value (tvb, valStart, &count);
						if (s) {
							ti = proto_tree_add_uint (tree, hf_wsp_header_max_forwards,
									tvb, 0, len, (*s));
							free (s);
						} else { /* Error: s */ }
					} else { /* Error */
						msg (MSG_WARN, "add_header: Max-Forwards: Invalid encoding (valType = %02X", valType);
					}
					break;

				case WSP_HDR_BEARER_INDICATION:
					if ((valType & 0x80) || (valType < 0x1F)) {
						guint32	count	= 0;
						guint32	*s		= get_bnf_integer_value (tvb, valStart, &count);
						if (s) {
							ti = proto_tree_add_uint (tree, hf_wsp_header_bearer_indication,
									tvb, 0, len, (*s));
							free (s);
						} else { /* Error: s */ }
					} else { /* Error */
						msg (MSG_WARN, "add_header: Bearer-Indication: Invalid encoding (valType = %02X", valType);
					}
					break;

				/* Short-int */
				case WSP_HDR_PUSH_FLAG:
					if (valType & 0x80) {
						ti = proto_tree_add_uint (tree, hf_wsp_header_push_flag,
								tvb, 0, len, valType & 0x7F);
					} else { /* Error: invalid encoding */
						msg (MSG_WARN, "add_header: Push-Flag: Invalid encoding (valType = %02X", valType);
					}
					break;


				case WSP_HDR_PRAGMA:
					if (valType & 0x80) {
						if (valType == 0x80) { /* No-cache */
							ti = proto_tree_add_string (tree, hf_wsp_header_pragma,
									tvb, 0, len, "No-Cache");
						} else { /* Error: invalid type */
						}
					} else if (valType <= 0x1F) { /* General form: Parameter */
						guint32	count	= 0;
						char	*s		= get_parameter (tvb, valStart, &count);
						if (s) {
							ti = proto_tree_add_string (tree, hf_wsp_header_pragma,
									tvb, 0, len, s);
							free (s);
						}
					} else { /* Error: wrong encoding */
						msg (MSG_WARN, "add_header: Pragma: Invalid encoding (valType = %02X", valType);
					}
					break;
					

				case WSP_HDR_CONNECTION:
					if (valType & 0x80) { /* Short-integer */
						if (valType == 0x80) { /* Close */
							ti = proto_tree_add_string (tree, hf_wsp_header_connection,
									tvb, 0, len, "Close");
						} else { /* Invalid method! */
							msg (MSG_WARN, "add_header: Connection header incorrectly encoded.");
						}
					} else if (istokenchar (valType)) { /* Token-text */
						if (valString) {
							ti = proto_tree_add_string (tree, hf_wsp_header_connection,
									tvb, 0, len, valString);
						} else { /* Error */ }
					} else {
						msg (MSG_WARN, "add_header: Connection header incorrectly encoded.");
					}
					break;


				case WSP_HDR_PUBLIC:
					if (valType & 0x80) { /* Short-integer: method */
						if (valType >= 0xC0) { /* Valid method */
							ti = proto_tree_add_uint (tree, hf_wsp_header_public,
									tvb, 0, len, valType & 0x7F);
						} else { /* Invalid method! */
							ti = proto_tree_add_uint (tree, hf_wsp_header_public,
									tvb, 0, len, 0);
						}
					} else if (istokenchar (valType)) { /* Token-text: method */
						if (valString) {
							ti = proto_tree_add_string (tree, hf_wsp_header_public_str,
									tvb, 0, len, valString);
						} else { /* Error */ }
					} else {
						msg (MSG_WARN, "add_header: Public header incorrectly encoded.");
					}
					break;
					

				case WSP_HDR_VARY:
					if (valType & 0x80) { /* Short-integer: Well-known-header */
						ti = proto_tree_add_uint (tree, hf_wsp_header_vary,
								tvb, 0, len, valType & 0x7F);
					} else if (istokenchar (valType)) { /* Token-text: header-name */
						if (valString) {
							ti = proto_tree_add_string (tree, hf_wsp_header_vary_str,
									tvb, 0, len, valString);
						} else { /* Error */ }
					} else {
						msg (MSG_WARN, "add_header: Vary header incorrectly encoded.");
					}
					break;
					
					
				case WSP_HDR_TRAILER:
					if (valType & 0x80) { /* Short-integer: Well-known-header */
						ti = proto_tree_add_uint (tree, hf_wsp_header_vary,
								tvb, 0, len, valType & 0x7F);
					} else if (valType < 0x1F) { /* Long-integer: idem */
						/* Only 16 bit can be decoded: len + (2 bit) + 7 bit + 7 bit */
						guint32	count	= 0;
						guint32	*s		= get_bnf_long_integer (tvb, valStart, &count);
						if (s) {
							if (count < 5) { /* We can decode it */
								ti = proto_tree_add_uint (tree, hf_wsp_header_trailer,
										tvb, 0, len, (*s));
							} else { /* Error: not decodable */
								msg (MSG_WARN, "Trailers: integer > 32 bits not decoded");
							}
							free (s);
						} else { /* Error: s */ }
					} else if (istokenchar (valType)) { /* Token-text: header-name */
						if (valString) {
							ti = proto_tree_add_string (tree, hf_wsp_header_trailer_str,
									tvb, 0, len, valString);
						} else { /* Error */ }
					} else {
						msg (MSG_WARN, "add_header: Trailer header incorrectly encoded.");
					}
					break;
					
					
				case WSP_HDR_TRANSFER_ENCODING:
					if (valType & 0x80) { /* Short-integer */
						if (valType == 0x80) { /* Valid method */
							ti = proto_tree_add_string (tree, hf_wsp_header_transfer_encoding_str,
									tvb, 0, len, "Chunked");
						} else { /* Invalid encoding */
							msg (MSG_WARN, "add_header: Transfer-Encoding header incorrectly encoded.");
						}
					} else if (istokenchar (valType)) { /* Token-text */
						if (valString) {
							ti = proto_tree_add_string (tree, hf_wsp_header_transfer_encoding_str,
									tvb, 0, len, valString);
						} else { /* Error */ }
					} else {
						msg (MSG_WARN, "add_header: Transfer-Encoding header incorrectly encoded.");
					}
					break;

				case WSP_HDR_AUTHORIZATION:
					/* 0x80 == Basic ; User-id = token-text; Pass = Token-text
					 * OR
					 * Token-text; *parameter
					 */
					if (valType <= 0x1F) { /* Length */
						guint32	off		= valStart;
						guint8	peek	= tvb_get_guint8 (tvb, off);
						if (peek & 0x80) { /* Short-integer */
							if (peek == 0x80) { /* Basic */
								off++;
								/* Token-text User-id */
								if (off < len) {
									peek = tvb_get_guint8 (tvb, off);
									if (istokenchar (peek)) {
										guint32	count	= 0;
										char	*userid = get_bnf_token_text (tvb, off, &count);
										off += count;
										if (userid) {
											/* Token-text Pass */
											if (off < len) {
												peek = tvb_get_guint8 (tvb, off);
												if (istokenchar (peek)) {
													guint32	count	= 0;
													char	*passwd = get_bnf_token_text (tvb, off, &count);
													off += count;
													if (passwd) {
														char *s = malloc (30 + off - valStart);
														if (s) {
															sprintf (s, "Basic (user='%s', password='%s')",
																	userid, passwd);
															ti = proto_tree_add_string (tree,
																	hf_wsp_header_authorization,
																	tvb, 0, len, s);
														} else { /* Error: malloc */ }
													} else { /* Error: malloc */ }
												} else { /* Error: encoding */
													msg (MSG_WARN, "Authorization: wrong encoding (expecting passwd: token-text)");
												}
											} else { /* Error: encoding: too small */
													msg (MSG_WARN, "Authorization: wrong encoding");
											}
										} else { /* Error: malloc */ }
									} else { /* Error: encoding */
										msg (MSG_WARN, "Authorization: wrong encoding (expecting userid: token-text)");
									}
								} else { /* Error: encoding: too small */
									msg (MSG_WARN, "Authorization: wrong encoding");
								}
							} else { /* Error: encoding */
								msg (MSG_WARN, "Authorization: wrong encoding");
							}
						} else if (istokenchar (peek)) { /* Token-text */
							guint32	count	= 0;
							guint32	off		= 0;
							char	*s		= get_bnf_token_text (tvb, off, &count);
							off				= valStart + count;
							ti = proto_tree_add_string (tree,
									hf_wsp_header_authorization,
									tvb, 0, len, s);
							/* Optional parameters */
							while (off < len) {
								off = add_parameter (tree, tvb, off);
							}
						} else { /* Error: encoding */
							msg (MSG_WARN, "Authorization: wrong encoding (expecting userid: token-text)");
						}
					} else { /* Error: encoding */
						msg (MSG_WARN, "Authorization: wrong encoding");
					}
					break;

				case WSP_HDR_PROXY_AUTHORIZATION:
					/* 0x80 == Basic ; User-id = token-text; Pass = Token-text
					 * OR
					 * Token-text; *parameter
					 */
					if (valType <= 0x1F) { /* Length */
						guint32	off		= valStart;
						guint8	peek	= tvb_get_guint8 (tvb, off);
						if (peek & 0x80) { /* Short-integer */
							if (peek == 0x80) { /* Basic */
								off++;
								/* Token-text User-id */
								if (off < len) {
									peek = tvb_get_guint8 (tvb, off);
									if (istokenchar (peek)) {
										guint32	count	= 0;
										char	*userid = get_bnf_token_text (tvb, off, &count);
										off += count;
										if (userid) {
											/* Token-text Pass */
											if (off < len) {
												peek = tvb_get_guint8 (tvb, off);
												if (istokenchar (peek)) {
													guint32	count	= 0;
													char	*passwd = get_bnf_token_text (tvb, off, &count);
													off += count;
													if (passwd) {
														char *s = malloc (30 + off - valStart);
														if (s) {
															sprintf (s, "Basic (user='%s', password='%s')",
																	userid, passwd);
															ti = proto_tree_add_string (tree,
																	hf_wsp_header_proxy_authorization,
																	tvb, 0, len, s);
														} else { /* Error: malloc */ }
													} else { /* Error: malloc */ }
												} else { /* Error: encoding */
													msg (MSG_WARN, "Proxy-Authorization: wrong encoding (expecting passwd: token-text)");
												}
											} else { /* Error: encoding: too small */
													msg (MSG_WARN, "Proxy-Authorization: wrong encoding");
											}
										} else { /* Error: malloc */ }
									} else { /* Error: encoding */
										msg (MSG_WARN, "Proxy-Authorization: wrong encoding (expecting userid: token-text)");
									}
								} else { /* Error: encoding: too small */
									msg (MSG_WARN, "Proxy-Authorization: wrong encoding");
								}
							} else { /* Error: encoding */
								msg (MSG_WARN, "Proxy-Authorization: wrong encoding");
							}
						} else if (istokenchar (peek)) { /* Token-text */
							guint32	count	= 0;
							char	*s		= get_bnf_token_text (tvb, off, &count);
							ti = proto_tree_add_string (tree,
									hf_wsp_header_proxy_authorization,
									tvb, 0, len, s);
							/* Optional parameters */
							while (off < len) {
								off = add_parameter (tree, tvb, off);
							}
						} else { /* Error: encoding */
							msg (MSG_WARN, "Proxy-Authorization: wrong encoding (expecting userid: token-text)");
						}
					} else { /* Error: encoding */
						msg (MSG_WARN, "Proxy-Authorization: wrong encoding");
					}
					break;


				case WSP_HDR_WWW_AUTHENTICATE:
					/* 0x80 == Basic ; Realm-value = Text-string
					 * OR
					 * Token-text; Realm-value = Text-string ; *parameter
					 */
					msg (MSG_WARN, "WWW-Authenticate: header being decoded (not yet checked)");
					if (valType <= 0x1F) { /* Length */
						guint32	off		= valStart;
						guint8	peek	= tvb_get_guint8 (tvb, off);
						if (peek & 0x80) { /* Short-integer */
							if (peek == 0x80) { /* Basic */
								off++;
								/* Text-string Realm-value */
								if (off < len) {
									peek = tvb_get_guint8 (tvb, off);
									if ((peek > 0x1F) && (peek < 0x80)) { /* Text-string */
										guint32	count	= 0;
										char	*realm = get_bnf_text_string (tvb, off, &count);
										off += count;
										if (realm) {
											char *s = malloc (25 + off - valStart);
											if (s) {
												sprintf (s, "Basic (Realm name = '%s')",
														realm);
												ti = proto_tree_add_string (tree,
														hf_wsp_header_www_authenticate,
														tvb, 0, len, s);
											} else { /* Error: malloc */ }
										} else { /* Error: malloc */ }
									} else { /* Error: encoding */
										msg (MSG_WARN, "WWW-Authenticate: wrong encoding (expecting realm-value: text-string)");
									}
								} else { /* Error: encoding: too small */
									msg (MSG_WARN, "WWW-Authenticate: wrong encoding");
								}
							} else { /* Error: encoding */
								msg (MSG_WARN, "WWW-Authenticate: wrong encoding");
							}
						} else if (istokenchar (peek)) { /* Token-text */
							guint32	count		= 0;
							char	*realm_type	= get_bnf_token_text (tvb, off, &count);
							off					+= count; 
							if (realm_type) {
								/* Text-string Realm-value */
								if (off < len) {
									peek = tvb_get_guint8 (tvb, off);
									if ((peek > 0x1F) && (peek < 0x80)) { /* Text-string */
										guint32	count	= 0;
										char	*realm = get_bnf_text_string (tvb, off, &count);
										off += count;
										if (realm) {
											char *s = malloc (20 + off - valStart);
											if (s) {
												sprintf (s, "%s (Realm name = '%s')",
														realm_type, realm);
												ti = proto_tree_add_string (tree,
														hf_wsp_header_www_authenticate,
														tvb, 0, len, s);
												/* Optional parameters */
												while (off < len) {
													off = add_parameter (tree, tvb, off);
												}
											} else { /* Error: malloc */ }
										} else { /* Error: malloc */ }
									} else { /* Error: encoding */
										msg (MSG_WARN, "WWW-Authenticate: wrong encoding (expecting realm-value: text-string)");
									}
								} else { /* Error: encoding: too small */
									msg (MSG_WARN, "WWW-Authenticate: wrong encoding");
								}
							} else { /* Error: malloc */ }
						} else { /* Error: encoding */
							msg (MSG_WARN, "WWW-Authenticate: wrong encoding (expecting userid: token-text)");
						}
					} else { /* Error: encoding */
						msg (MSG_WARN, "WWW-Authenticate: wrong encoding");
					}
					break;
				case WSP_HDR_PROXY_AUTHENTICATE:



				default:
					if (hdrType & 0x80) { /* Well-known */
						if (hdrID != 0xFF) {
							ti = proto_tree_add_text (tree, tvb, 0, len,
									"%s: (not yet parsed)", match_strval (hdrID, vals_header_names));
							msg (MSG_WARN, "Well-known header 0x%02X: not yet parsed", hdrType & 0x7F);
						} else {
							ti = proto_tree_add_text (tree, tvb, 0, len,
									"Unknown Header (ID 0x%02X)", hdrType & 0x7F);
							msg (MSG_WARN, "Unknown header 0x%02X: not parsed", hdrType & 0x7F);
						}
					} else { /* Textual */
							if (valType <= 0x1F) {
								msg (MSG_WARN, "Textual header '%s': not known and not parsed", hdrString);
								ti = proto_tree_add_text (tree, tvb, 0, len,
										"%s: (header value not decoded)", hdrString);
							} else if (valType < 0x80) {
								guint count = 0;
								msg (MSG_WARN, "Textual header '%s': not known and not parsed", hdrString);
								valString	= (char *)tvb_get_ptr (tvb, valStart, len - valStart);
								ti = proto_tree_add_text (tree, tvb, 0, len,
										"%s: %s", hdrString, get_bnf_text_string (tvb, valStart, &count));
							} else {
								msg (MSG_WARN, "Textual header '%s': not known and not parsed", hdrString);
								ti = proto_tree_add_text (tree, tvb, 0, len,
										"%s: (header value 0x%02X not decoded)", hdrString, valType);
							}
					}
					break;
			}
			break;	/* Header code page 1 */

		default:
			if (hdrType & 0x80) { /* Well-known */
				if (valType <= 0x1F) {
					ti = proto_tree_add_text (tree, tvb, 0, len,
							"HCP 0x%02X: Known-Header 0x%02X: %d bytes of value data (not parsed)",
							curHCP, hdrType & 0x7F, valLen);
				} else if (valType < 0x80) {
					ti = proto_tree_add_text (tree, tvb, 0, len,
							"HCP 0x%02X: Known-Header 0x%02X: '%s'",
							curHCP, hdrType & 0x7F, valString);
				} else {
					ti = proto_tree_add_text (tree, tvb, 0, len,
							"HCP 0x%02X: Known-Header 0x%02X: Known-Value 0x%02X (not parsed)",
							curHCP, hdrType & 0x7F, valType & 0x7F);
				}
			} else { /* Textual */
				if (valType <= 0x1F) {
					ti = proto_tree_add_text (tree, tvb, 0, len,
							"HCP 0x%02X: '%s': %d bytes of value data (not parsed)",
							curHCP, hdrString, valLen);
				} else if (valType < 0x80) {
					ti = proto_tree_add_text (tree, tvb, 0, len,
							"HCP 0x%02X: '%s': '%s'",
							curHCP, hdrString, valString);
				} else {
					ti = proto_tree_add_text (tree, tvb, 0, len,
							"HCP 0x%02X: '%s': Known-Value 0x%02X (not parsed)",
							curHCP, hdrString, valType & 0x7F);
				}
			}
			break;
	}
}





static guint
get_value_length (tvbuff_t *tvb, guint offset, guint *nextOffset)
{
	guint value = 0;
	guint count = 0;
	guint octet = tvb_get_guint8 (tvb, offset);

	if (octet <= 30)	/* Short length */
	{
		value = octet;
		*nextOffset = offset+1;
	}
	else if (octet == 31)
	{
		value = tvb_get_guintvar (tvb, offset+1, &count);
		*nextOffset = offset+1+count;
	}
	else
	{
		fprintf (stderr, "dissect_wsp: get_value_length: case NYI\n");
	}

	return (value);
}


static void
add_redirect_addresses (proto_tree *tree, tvbuff_t *tvb)
{
	proto_item *ti;
	proto_tree *wsp_redirect_addresses;
	guint offset = 0;
	guint Len = tvb_reported_length (tvb);
	guint8 Start = 0;
	guint peek = 0;

	msg (MSG_INFO, "add_redirect_addresses: Offset is %d, size is %d", offset, Len);

	/* End of buffer */
	if (Len <= 0)
	{
		return;
	}

	msg (MSG_INFO, "Redirect address(es) to process");

	ti = proto_tree_add_item (tree, hf_wsp_redirect_addresses_section, tvb,
			0, Len, bo_big_endian);
	
	wsp_redirect_addresses = proto_item_add_subtree( ti, ett_redirect );

	/* Parse Redirect Addresses */

	while (offset < Len)
	{
		char		bearerTypeFlag,
					portNumberFlag,
					addressLength,
					recordLength	= 1;
		guint8		bearerType		= 0;	/* Only used when bearerTypeFlag is set */
		proto_tree	*address_rec_tree;
		
		Start				= offset;
		peek				= tvb_get_guint8 (tvb, Start);
		bearerTypeFlag		= peek & 0x80;
		portNumberFlag		= peek & 0x40;
		addressLength		= peek & 0X3F;
		
		if (bearerTypeFlag) {
			bearerType = tvb_get_guint8 (tvb, Start + 1);
			recordLength	++;
		}
		if (portNumberFlag)
			recordLength	+= 2;
		recordLength		+= addressLength;
		
		msg (MSG_INFO, "Redirect: Address record length is %d (offset is %d).",
				recordLength, offset);

		ti = proto_tree_add_item (wsp_redirect_addresses, hf_wsp_redirect_address_section, tvb,
				offset, recordLength, bo_big_endian);
		address_rec_tree = proto_item_add_subtree (ti, ett_redir_addr);
		ti = proto_tree_add_boolean (address_rec_tree, hf_wsp_redirect_bearertype_included,
				tvb, offset, 1, bearerTypeFlag);
		ti = proto_tree_add_boolean (address_rec_tree, hf_wsp_redirect_port_included,
				tvb, offset, 1, portNumberFlag);
		ti = proto_tree_add_uint (address_rec_tree, hf_wsp_redirect_address_length,
				tvb, offset, 1, addressLength);
		offset++;
		
		if (peek & 0x80) { /* Bearer Type included */
			ti = proto_tree_add_uint (address_rec_tree, hf_wsp_redirect_bearertype,
					tvb, offset, 1, tvb_get_guint8 (tvb, offset));
			offset += 1;
		}
		if (peek & 0x40) { /* Port Number included */
			ti = proto_tree_add_uint (address_rec_tree, hf_wsp_redirect_port,
					tvb, offset, 2, tvb_get_ntohs (tvb, offset));
			offset += 2;
		}
		if (bearerTypeFlag) {
			if ((bearerType == 0)
			||	((0x06 <= bearerType) && (bearerType <= 0x0F))
			||	(bearerType == 0x11)
			||	(bearerType == 0x12)
			||	(bearerType == 0x15)
			||	(bearerType == 0x19)) { /* IPv4 address */
				ti = proto_tree_add_text (address_rec_tree, tvb,
						offset, addressLength, "Address: (IPv4) %d.%d.%d.%d",
						tvb_get_guint8(tvb,offset),			tvb_get_guint8(tvb,offset + 1),
						tvb_get_guint8(tvb,offset + 2),		tvb_get_guint8(tvb,offset + 3));
			} else if ( bearerType == 0x01) { /* IPv6 address */
				ti = proto_tree_add_text (address_rec_tree, tvb,
						offset, addressLength, "Address: (IPv6) "
						"%d:%d:%d:%d %d:%d:%d:%d %d:%d:%d:%d %d:%d:%d:%d",
						tvb_get_guint8(tvb,offset),			tvb_get_guint8(tvb,offset + 1),
						tvb_get_guint8(tvb,offset + 2),		tvb_get_guint8(tvb,offset + 3),
						tvb_get_guint8(tvb,offset + 4),		tvb_get_guint8(tvb,offset + 5),
						tvb_get_guint8(tvb,offset + 6),		tvb_get_guint8(tvb,offset + 7),
						tvb_get_guint8(tvb,offset + 8),		tvb_get_guint8(tvb,offset + 9),
						tvb_get_guint8(tvb,offset + 10),	tvb_get_guint8(tvb,offset + 11),
						tvb_get_guint8(tvb,offset + 12),	tvb_get_guint8(tvb,offset + 13),
						tvb_get_guint8(tvb,offset + 14),	tvb_get_guint8(tvb,offset + 15));
			} else { /* Other addresses */
				ti = proto_tree_add_text (address_rec_tree, tvb,
						offset, addressLength, "Address: (not yet decoded)");
			}
			offset += addressLength;
		} else { /* Look up in protocol tree for knowing which protocol is carrying WDP */
			ti = proto_tree_add_text (address_rec_tree, tvb,
					offset, addressLength, "Address: (same type as from this packet - not yet decoded)");
		}
	} /* END while */
}


static void
add_capabilities (proto_tree *tree, tvbuff_t *tvb, char pdut)
{
	proto_item	*ti;
	proto_tree	*capa_tree;
	guint		offset	= 0;
	guint		Len		= tvb_reported_length (tvb);

	/* End of buffer */
	if (Len <= 0)
	{
		return;
	}

	msg (MSG_INFO, "Capabilities to process");

	ti = proto_tree_add_item (tree, hf_wsp_capabilities_section, tvb,
			0, Len, bo_big_endian);
	
	capa_tree = proto_item_add_subtree (ti, ett_capabilities);
	
	/* Capability decoding rules
	 * =========================
	 * Length       : guintvar (1-4 OCTETs)
	 * Identifier   : Token-text | Short-integer
	 * Parameter[s] : Depending on Identifier (and PDU type)
	 */
	while (offset < Len) {
		guint32		count		= 0,
					capaStart	= offset,	/* Start of a capability record */
					fullCapaLen	= 0,		/* Length of a capability record including length field */
					capaLen		= 0,		/* Length of a capability record without length field */
					capaType	= 0,
					capaTypeLen	= 0,		/* Length of capability identifier */
					capaValLen	= 0;		/* Length of caoability value */
		proto_tree	*capa_rec_tree	= NULL;

		/* Read length of this capability */
		capaLen		= tvb_get_guintvar(tvb, offset, &count);
		fullCapaLen	= count + capaLen;
		offset		+= count;
		
		/* Read type of this capablility */
		capaType = tvb_get_guint8 (tvb, offset);
		
		msg (MSG_INFO, "CapaType = 0x%02X, offset = %d, length = %d.",
				capaType, offset, capaLen);
		
		capa_rec_tree = capa_tree;
		if (capaType & 0x80) { /* well-known-type */
			capaTypeLen = 1;
		} else {
			char	*s	= (char *)tvb_get_ptr(tvb, offset, capaLen);
			ti = proto_tree_add_text (capa_tree, tvb,
					offset, capaLen + count,
					"%s : (Textual capability not decoded)", (s ? s : "") );
			offset += count;
			capaTypeLen = (s ? strlen(s) + 1 : 1);
		}
		offset += capaTypeLen;
		capaValLen = capaLen - capaTypeLen;
		
		/* Read value of this capability */
		if (capaTypeLen == 1) { /* Known capability type */
			switch (capaType & 0x7F) {
				case 0: /* Client SDU size */
					ti = proto_tree_add_uint (capa_rec_tree,
							hf_wsp_capability_client_sdu_size,
							tvb, capaStart, fullCapaLen,
							tvb_get_guintvar(tvb, offset, &count));
					offset += count;
					break;
				
				case 1: /* Server SDU size */
					ti = proto_tree_add_uint (capa_rec_tree,
							hf_wsp_capability_server_sdu_size,
							tvb, capaStart, fullCapaLen,
							tvb_get_guintvar(tvb, offset, &count));
					offset += count;
					break;
					
				case 2: /* Protocol options */
					ti = proto_tree_add_item (capa_tree, hf_wsp_capability_proto_options, tvb,
							capaStart, fullCapaLen, bo_big_endian);
					/* As of WAP-203-WSP (June 20, 2000 release) this field is only 1 byte */
					if (capaValLen == 1) {
						guint8	flags = tvb_get_guint8(tvb, offset);
						int		i;
						
						for (i = 0x01; i <= 0x80; i <<= 1) { /* Display the protocol options */
							if (flags & i) {
								ti = proto_tree_add_uint (capa_rec_tree,
										hf_wsp_capability_proto_option,
										tvb, offset, 1, i);
							}
						}
					} else {
						msg (MSG_WARN, "Capability: Protocol Options: "
								"multi-octet flags not yet defined/supported");
						offset += fullCapaLen;
					}
					break;
					
				case 3: /* Method MOR */
					ti = proto_tree_add_uint (capa_rec_tree, hf_wsp_capability_method_mor,
							tvb, capaStart, fullCapaLen, tvb_get_guint8(tvb, offset));
					break;
					
				case 4: /* Push MOR */
					ti = proto_tree_add_uint (capa_rec_tree, hf_wsp_capability_push_mor,
							tvb, capaStart, fullCapaLen, tvb_get_guint8(tvb, offset));
					break;
					
				case 5: /* Extended methods */
					/* First read byte for short-hand, then read null-terminated string */
					{	guint	start	= offset,
								nameLen	= 0,
								end		= capaStart + fullCapaLen;
						char	*name = NULL;
						
						ti = proto_tree_add_item (capa_rec_tree, hf_wsp_capability_xmethods,
								tvb, capaStart, fullCapaLen, bo_big_endian);
						while ( start < end ) {
							guint8 method = tvb_get_guint8 (tvb, start);
							start++;
							if (pdut == CONNECT) {
								char	*s;
								
								name	= (char *)tvb_get_ptr (tvb, start, end - start);
								nameLen	= (name ? strlen (name) + 1 : 1);
								s		= malloc ((7 + nameLen) * sizeof (char));
								if (s) {
									sprintf (s, "0x%02X = %s", method, name);
									if ((0x50 <= method) && (method <= 0x5F)) {
										ti = proto_tree_add_string (capa_rec_tree,
												hf_wsp_capability_xmethod_get_val,
												tvb, start - 1, 1 + nameLen, s);
									} else if ((0x70 <= method) && (method <= 0x7F)) {
										ti = proto_tree_add_string (capa_rec_tree,
												hf_wsp_capability_xmethod_push_val,
												tvb, start - 1, 1 + nameLen, s);
									} else {
										msg (MSG_WARN, "Capability: "
												"Extended Methods: illegal method 0x%02X", method);
									}
								} else {
									msg (MSG_ERROR, "Capability: "
											"Extended Methods: malloc () failed.");
									start = end;
								}
								start += nameLen;
							} else {
								if ((0x50 <= method) && (method <= 0x5F)) {
									ti = proto_tree_add_uint (capa_rec_tree,
											hf_wsp_capability_xmethod_get,
											tvb, start - 1, 1, method);
								} else if ((0x70 <= method) && (method <= 0x7F)) {
									ti = proto_tree_add_uint (capa_rec_tree,
											hf_wsp_capability_xmethod_push,
											tvb, start - 1, 1, method);
								} else {
									msg (MSG_WARN, "Capability: "
											"Extended Methods: illegal method 0x%02X", method);
								}
							}
						}
					}
					break;
				case 6: /* Header code pages */
					/* First read byte for short-hand, then read optional null-terminated string */
					{	guint	start	= offset,
								nameLen	= 0,
								end		= capaStart + fullCapaLen;
						char	*name = NULL;
						
						ti = proto_tree_add_item (capa_rec_tree,
								hf_wsp_capability_header_code_pages,
								tvb, capaStart, fullCapaLen, bo_big_endian);
						while ( start < end ) {
							guint8 hcp = tvb_get_guint8 (tvb, start);
							start++;
							if (pdut == CONNECT) {
								char *s;

								name	= (char *)tvb_get_ptr (tvb, start, end - start);
								nameLen	= (name ? strlen (name) + 1 : 1);
								s		= malloc ((7 + nameLen) * sizeof (char));
								if (s) {
									sprintf (s, "0x%02X = %s", hcp, name);
									ti = proto_tree_add_string (capa_rec_tree,
											hf_wsp_capability_header_code_page_val,
											tvb, start - 1, 1 + nameLen, s);
								} else {
									msg (MSG_ERROR, "Capability: "
											"Header Code Pages: malloc () failed.");
									start = end;
								}
								start += nameLen;
							} else {
								ti = proto_tree_add_uint (capa_rec_tree,
										hf_wsp_capability_header_code_page,
										tvb, start - 1, 1, hcp);
							}
						}
					}
					break;
				case 7: /* Aliases */
					{
						guint	start	= offset,
								end		= offset + capaLen;
						/* Parse Alias Addresses */

						while (start < end)
						{
							char	bearerTypeFlag,
									portNumberFlag,
									addressLength,
									recordLength	= 1;
							guint8	bearerType		= 0;	/* Only used when bearerTypeFlag is set */
							guint8	peek			= tvb_get_guint8 (tvb, start);

							bearerTypeFlag		= peek & 0x80;
							portNumberFlag		= peek & 0x40;
							addressLength		= peek & 0X3F;
		
							if (bearerTypeFlag) {
								bearerType = tvb_get_guint8 (tvb, start + 1);
								recordLength	++;
							}
							if (portNumberFlag)
								recordLength	+= 2;
							recordLength		+= addressLength;
							
							msg (MSG_INFO, "Alias: Address record length is %d (offset is %d).",
									recordLength, start);
							
							ti = proto_tree_add_item (capa_rec_tree,
									hf_wsp_capability_alias_address_section,
									tvb, start, recordLength, bo_big_endian);
							ti = proto_tree_add_boolean (capa_rec_tree,
									hf_wsp_capability_alias_bearertype_included,
									tvb, start, 1, bearerTypeFlag);
							ti = proto_tree_add_boolean (capa_rec_tree,
									hf_wsp_capability_alias_port_included,
									tvb, start, 1, portNumberFlag);
							ti = proto_tree_add_uint (capa_rec_tree,
									hf_wsp_capability_alias_address_length,
									tvb, start, 1, addressLength);
							start++;
			
							if (peek & 0x80) { /* Bearer Type included */
								ti = proto_tree_add_uint (capa_rec_tree,
										hf_wsp_capability_alias_bearertype,
										tvb, start, 1, tvb_get_guint8 (tvb, start));
								start += 1;
							}
							if (peek & 0x40) { /* Port Number included */
								ti = proto_tree_add_uint (capa_rec_tree,
										hf_wsp_capability_alias_port,
										tvb, start, 2, tvb_get_ntohs (tvb, start));
								start += 2;
							}
							if (bearerTypeFlag) {
								if ((bearerType == 0)
								||	((0x06 <= bearerType) && (bearerType <= 0x0F))
								||	(bearerType == 0x11)
								||	(bearerType == 0x12)
								||	(bearerType == 0x15)
								||	(bearerType == 0x19)) { /* IPv4 address */
									ti = proto_tree_add_text (capa_rec_tree, tvb,
											start, addressLength, "Address: (IPv4) %d.%d.%d.%d",
											tvb_get_guint8(tvb,start),		tvb_get_guint8(tvb,start + 1),
											tvb_get_guint8(tvb,start + 2),	tvb_get_guint8(tvb,start + 3));
								} else if ( bearerType == 0x01) { /* IPv6 address */
									ti = proto_tree_add_text (capa_rec_tree, tvb,
											start, addressLength, "Address: (IPv6) "
											"%d:%d:%d:%d %d:%d:%d:%d %d:%d:%d:%d %d:%d:%d:%d",
											tvb_get_guint8(tvb,start),		tvb_get_guint8(tvb,start + 1),
											tvb_get_guint8(tvb,start + 2),	tvb_get_guint8(tvb,start + 3),
											tvb_get_guint8(tvb,start + 4),	tvb_get_guint8(tvb,start + 5),
											tvb_get_guint8(tvb,start + 6),	tvb_get_guint8(tvb,start + 7),
											tvb_get_guint8(tvb,start + 8),	tvb_get_guint8(tvb,start + 9),
											tvb_get_guint8(tvb,start + 10),	tvb_get_guint8(tvb,start + 11),
											tvb_get_guint8(tvb,start + 12),	tvb_get_guint8(tvb,start + 13),
											tvb_get_guint8(tvb,start + 14),	tvb_get_guint8(tvb,start + 15));
								} else { /* Other addresses */
									ti = proto_tree_add_text (capa_rec_tree, tvb,
											start, addressLength, "Address: (not yet decoded)");
								}
								start += addressLength;
							} else { /* Look up in protocol tree for knowing which protocol is carrying WDP */
								ti = proto_tree_add_text (capa_rec_tree, tvb,
										start, addressLength, "Address: (same type as from this packet - not yet decoded)");
								start += addressLength;
							}
						} /* END while */
					}
					break;
				default: /* Unknown well-known capability */
					msg (MSG_WARN, "Unknown Well-known-capability 0x%02X (skipped)",
							capaType & 0x7F);
					offset += capaLen;	/* Skip this capability */
					break;
			}
		} else { /* Textual capability */
			msg (MSG_WARN, "Textual capability (not decoded)",
					capaType & 0x7F);
			offset += capaLen;	/* Skip this capability */
		}
		offset = capaStart + fullCapaLen;
	}
}



static guint
add_content_type (proto_tree *tree, tvbuff_t *tvb, guint offset, guint *contentType)
{
	proto_tree *contentTypeTree;
	guint nextOffset = offset;
	guint fieldLength = 0;
	guint octet = tvb_get_guint8 (tvb, offset);
	guint totalSizeOfField = 0;

	if (! octet) /* No content-type field (error code PDU, ...) */
		return offset + 1;

	if (octet <= 31)
	{
		fieldLength = get_value_length (tvb, offset, &nextOffset);
		totalSizeOfField = (nextOffset-offset)+fieldLength;
	}
	else if (octet & 0x80)
	{
		fieldLength = 1;
		totalSizeOfField = 1;
	}
	else
	{
		fprintf (stderr, "dissect-wsp: Content-type is un-supported\n");
	}

	*contentType = (tvb_get_guint8 (tvb, nextOffset) & 0x7F);
	contentTypeTree = proto_tree_add_uint (tree, hf_wsp_content_type,
			tvb, offset, totalSizeOfField, (tvb_get_guint8(tvb,nextOffset++) & 0x7F));

	while (nextOffset < (offset+totalSizeOfField))
	{
		/* add_parameter */
		msg (MSG_INFO, "add_content_type: Before: parameter at offset %d, field length %d.",
				nextOffset, totalSizeOfField);
		nextOffset = add_parameter (contentTypeTree, tvb, nextOffset);
		msg (MSG_INFO, "add_content_type: After: parameter at offset %d, field length %d.",
				nextOffset, totalSizeOfField);
	}

	return (offset+totalSizeOfField);
}



static guint
add_parameter_q (proto_tree *tree, tvbuff_t *tvb, guint offset)
{
	guint	count = 0;
	guint32	*s	= get_bnf_q_value (tvb, offset, &count);
	if (s) {
		proto_tree_add_text (tree, tvb, offset, count,
				"Q = '0.%03d'", *s);
		free (s);
	} else { /* Error: s */ }
	return (offset + count + 1);
}



static char *
get_parameter (tvbuff_t *tvb, guint32 offset, guint32 *cnt)
{
	char	*s		= NULL;
	char	*r		= NULL;
	guint	count	= 0;
	guint	octet	= tvb_get_guint8 (tvb, offset);
	guint32	len		= tvb_reported_length (tvb);
	
	if (octet & 0x80)	/* Short integer */
	{
		offset++;
		(*cnt)++; /* parameter_type */
		octet &= 0x7F;
		msg (MSG_INFO, "add_parameter (type 0x%02X)", octet);
		switch ( octet )
		{
			case 0x00: /* Q */
				{
				guint32	*s	= get_bnf_q_value (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc (12 * sizeof (char));
					if (r) {
						sprintf (r, "Q = '0.%03d'", *s);
						free (s);
						return r;
					} else { /* Error: ENOMEM */
						return NULL;
					}
				} else { /* Error: s */
					offset	= len;
					(*cnt)	= len;
				}
				}
				break;
			case 0x01: /* Charset */
				{
				guint8	peek	= tvb_get_guint8 (tvb, offset);
				/* Short-int OR Long-int OR Token-text */
				if (peek & 0x80) { /* Short-integer */
					char *r = match_strval(peek & 0X7F, vals_character_sets);
					offset++;
					(*cnt)++;
					if (r) { /* Match */
						char *v = malloc ((strlen (r) + 1 + 13) * sizeof (char));
						if (v) {
							sprintf (v, "Charset = '%s'", r);
							return v;
						} else { /* Error: ENOMEM */
							return NULL;
						}
					} else { /* No match */
						/* 2^8 = 256 => 3 chars */
						char *v = malloc ((3 + 16 + 1 + 13) * sizeof (char));
						if (v) {
							sprintf (v, "Charset = Known-charset 0x%02X", (*r) & 0x7F);
							return v;
						} else { /* Error: ENOMEM */
							return NULL;
						}
					}
				} else if (peek < 0x1F) { /* Long-integer */
					guint	count	= 0;
					guint32	*s		= get_bnf_long_integer (tvb, offset, &count);
					if (s) {
						offset += count;
						(*cnt) += count;
						if (count <= 4) { /* 24 bit max */
							char *r = match_strval(*s, vals_character_sets);
							if (r) { /* Match */
								char *v = malloc ((strlen (r) + 1 + 13) * sizeof (char));
								if (v) {
									sprintf (v, "Charset = '%s'", r);
									free (s); /* No free of r !!! */
									return v;
								} else { /* Error: ENOMEM */ }
							} else { /* No match */
								/* 2^32 ~ 4 * 10^6 => 7 chars */
								char *v = malloc ((7 + 1 + 13) * sizeof (char));
								if (v) {
									sprintf (v, "Charset = '%d'", (*s));
									free (s); /* No free of r !!! */
									return v;
								} else { /* Error: ENOMEM */ }
							}
						} else { /* Too long for 24-bit int */
							msg (MSG_WARN, "Charset: value (> 24 bit) not decoded");
						}
						free (s);
						offset += count;
						(*cnt) += count;
					} else { /* Error: s */
						offset = len;	/* Skip further processing */
					}
				} else if (istokenchar (peek)) { /* Token-text */
					char *s	= get_bnf_token_text (tvb, offset, &count);
					(*cnt) += count;
					if (s) {
						char *r = malloc ((count + 1 + 13) * sizeof (char));
						if (r) {
							sprintf (r, "Charset = '%s'", s);
							free (s);
							return r;
						} else { /* Error: ENOMEM */ }
						free (s);
					} else { /* Error: s */ }
				} else { /* Error: invalid coding of this capability */
					offset = len;	/* Skip further processing */
				}
				}
				break;
			case 0x02: /* Level */
				s	= get_bnf_version_value (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((strlen (s) + 11) * sizeof (char));
					if (r) {
						sprintf (r, "Level = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x03: /* Type */
				{
				guint32	*s	= get_bnf_integer_value (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					/* 32 bits ~ 4 * 10^6 = 7 chars */
					r = malloc ((9 + 8) * sizeof (char));
					if (r) {
						sprintf (r, "Type = '%d'", *s);
						free (s);
						return r;
					}
				}
				}
				break;
			case 0x05: /* Name */
				s	= get_bnf_text_string (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((count + 10) * sizeof (char));
					if (r) {
						sprintf (r, "Name = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x06: /* Filename */
				s	= get_bnf_text_string (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((count + 14) * sizeof (char));
					if (r) {
						sprintf (r, "Filename = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x07: /* Differences */
				/* Field-name := Well-known-field-name | Token-text
				 * Well-known-field-name := Short-integer
				 */
				{
				guint8 peek = tvb_get_guint8 (tvb, offset);
				if (peek & 0x80) { /* Well-known-header */
					(*cnt) ++;
					s = match_strval (peek & 0x7F, vals_header_names);
					if (s) {
						r = malloc ((strlen (s) + 1 + 17) * sizeof (char));
						if (r) {
							sprintf (r, "Differences = '%s'", s);
							free (s);
							return r;
						} else { /* Error: ENOMEM */ }
						free (s);
					} else { /* No match */
						r = malloc ((18 + 1 + 17) * sizeof (char));
						if (r) {
							sprintf (r, "Differences = Known-Header 0x%02d", peek & 0x7F);
							return r;
						} else { /* Error: ENOMEM */ }
					}
				} else if (istokenchar (peek)) {
					guint32	count	= 0;
					s		= get_bnf_text_string (tvb, offset, &count);
					(*cnt)	+= count;
					if (s) {
						r = malloc ((count + 1 + 17) * sizeof (char));
						if (r) {
							sprintf (r, "Differences = '%s'", s);
							free (s);
							return r;
						} else  { /* Error: ENOMEM */ }
						free (s);
					} else { /* Error: s */ }
				} else { /* Error: invalid coding */
					msg (MSG_WARN, "Differences: Invalid header coding (1st octet = %02X)", peek);
					/* (*cnt) += capaLen;	*/
					offset	= len;	/* Skip this capa */
					*cnt	= len;
				}
				}
				break;
			case 0x08: /* Padding */
				{
				guint32	*s	= get_bnf_short_integer (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((4 + 13) * sizeof (char));
					if (r) {
						sprintf (r, "Padding = '%u'", (*s));
						free (s);
						return r;
					}
				}
				}
				break;
			case 0x09: /* Type (Content-type: multipart/related) */
				s	= get_bnf_text_string (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((count + 30) * sizeof (char));
					if (r) {
						sprintf (r, "Type (multipart/related) = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x0A: /* Start (multipart/related) */
				s	= get_bnf_text_string (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((count + 31) * sizeof (char));
					if (r) {
						sprintf (r, "Start (multipart/related) = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x0B: /* Start-info (multipart/related) */
				s	= get_bnf_text_string (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((count + 36) * sizeof (char));
					if (r) {
						sprintf (r, "Start-info (multipart/related) = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x0C: /* Comment */
				s	= get_bnf_text_string (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((count + 13) * sizeof (char));
					if (r) {
						sprintf (r, "Comment = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x0D: /* Domain */
				s	= get_bnf_text_string (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((count + 12) * sizeof (char));
					if (r) {
						sprintf (r, "Domain = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x0E: /* Max-Age */
				{
				guint32	*s	= get_bnf_integer_value (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					if (count <= 5) {
						r = malloc ((7 + 13) * sizeof (char));
						if (r) {
							sprintf (r, "Max-Age = '%d'", *s);
							free (s);
							return r;
						} else { /* Error: ENOMEM */ }
					} else { /* Too big uint */ }
					free (s);
				} else { /* Error: s */ }
				}
				break;
			case 0x0F: /* Path */
				s	= get_bnf_text_string (tvb, offset, &count);
				(*cnt) += count;
				if (s) {
					r = malloc ((count + 10) * sizeof (char));
					if (r) {
						sprintf (r, "Path = '%s'", s);
						free (s);
						return r;
					}
				}
				break;
			case 0x10: /* Secure */
				(*cnt)++;
				/* No-value == 0x00 */
				if (tvb_get_guint8 (tvb, offset) == 0) {
					r = malloc (7 * sizeof (char));
					strcpy (r, "Secure");
					return r;
				}
				break;

			default:
				msg (MSG_WARN, "wsp: add_parameter octet=0x%02x: not yet implemented", octet);
		};
		msg (MSG_INFO, "add_parameter (type 0x%02X), count=%d, offset=%d, newOffset=%d.",
				octet, count, offset, *cnt);
	}
	else
	{
		msg (MSG_INFO, "wsp: add_parameter octet=0x%02x", octet);
	}

	return NULL;
}

static guint
add_parameter (proto_tree *tree, tvbuff_t *tvb, guint offset)
{
	guint32 count	= 0;
	guint32	len		= tvb_reported_length (tvb);
	char	*s		= NULL;
	
	if (offset >= len) {
		msg (MSG_WARN, "wsp: add_parameter: buffer overrun (offset = %d, len = %d)", offset, len);
		return len; /* This prevents further processing of parameters */
	}

	if ((s = get_parameter (tvb, offset, &count))) {
		proto_tree_add_text (tree, tvb, offset, count, s);
	} else { /* An error occurred */
		msg (MSG_WARN, "wsp: add_parameter: no parameter value returned!");
	}

	return (offset + count);
}



static void
add_post_data (proto_tree *tree, tvbuff_t *tvb, guint contentType)
{
	guint offset = 0;
	guint variableStart = 0;
	guint variableEnd = 0;
	guint valueStart = 0;
	guint valueEnd = 0;
	guint8 peek = 0;
	proto_item *ti;
	
	/* VERIFY ti = proto_tree_add_item (tree, hf_wsp_post_data,tvb,offset,tvb_length_remaining(tvb, offset),bo_big_endian); */
	ti = proto_tree_add_item (tree, hf_wsp_post_data,tvb,offset,tvb_reported_length(tvb),bo_big_endian);

	if (contentType == 0x12)	/* URL Encoded data */
	{
		/* Iterate through post data */
		for (offset = 0; offset < tvb_reported_length (tvb); offset++)
		{
			peek = tvb_get_guint8 (tvb, offset);
			if (peek == '=')
			{
				variableEnd = offset;
				valueStart = offset+1;
			}
			else if (peek == '&')
			{
				if (variableEnd > 0)
				{
					add_post_variable (ti, tvb, variableStart, variableEnd, valueStart, offset);
				}
				variableStart = offset+1;
				variableEnd = 0;
				valueStart = 0;
				valueEnd = 0;
			}
		}

		/* See if there's outstanding data */
		if (variableEnd > 0)
		{
			add_post_variable (ti, tvb, variableStart, variableEnd, valueStart, offset);
		}
	}
}

static void
add_post_variable (proto_tree *tree, tvbuff_t *tvb, guint variableStart, guint variableEnd, guint valueStart, guint valueEnd)
{
	int variableLength = variableEnd-variableStart;
	int valueLength = 0;
	char *variableBuffer;
	char *valueBuffer;

	variableBuffer = g_malloc (variableLength+1);
	strncpy (variableBuffer, tvb_get_ptr (tvb, variableStart, variableLength), variableLength);
	variableBuffer[variableLength] = 0;

	if (valueEnd < valueStart)
	{
		valueBuffer = g_malloc (1);
		valueBuffer[0] = 0;
		valueEnd = valueStart;
	}
	else
	{
		valueLength = valueEnd-valueStart;
		valueBuffer = g_malloc (valueLength+1);
		strncpy (valueBuffer, tvb_get_ptr (tvb, valueStart, valueLength), valueLength);
		valueBuffer[valueLength] = 0;
	}

	/* Check for variables with no value */
	if (valueStart >= tvb_reported_length (tvb))
	{
		valueStart = tvb_reported_length (tvb);
		valueEnd = valueStart;
	}
	valueLength = valueEnd-valueStart;

	proto_tree_add_text (tree, tvb, variableStart, valueEnd-variableStart, "%s: %s", variableBuffer, valueBuffer);

	g_free (variableBuffer);
	g_free (valueBuffer);
}

/* Register the protocol with Ethereal */
void
proto_register_wsp(void)
{                 

	/* Setup list of header fields */
	static hf_register_info hf[] = {
		{ &hf_wsp_header_tid,
			{ 	"Transmission ID",           
				"wsp.TID",
				 FT_UINT8, BASE_HEX, NULL, 0x00,
				"Transmission ID" 
			}
		},
		{ &hf_wsp_header_pdu_type,
			{ 	"PDU Type",           
				"wsp.pdu_type",
				 FT_UINT8, BASE_HEX, VALS( vals_pdu_type ), 0x00,
				"PDU Type" 
			}
		},
		{ &hf_wsp_version_major,
			{ 	"Version (Major)",           
				"wsp.version.major",
				 FT_UINT8, BASE_DEC, NULL, 0xF0,
				"Version (Major)" 
			}
		},
		{ &hf_wsp_version_minor,
			{ 	"Version (Minor)",           
				"wsp.version.minor",
				 FT_UINT8, BASE_DEC, NULL, 0x0F,
				"Version (Minor)" 
			}
		},
		{ &hf_wsp_capability_length,
			{ 	"Capability Length",           
				"wsp.capability.length",
				 FT_UINT32, BASE_DEC, NULL, 0x00,
				"Capability Length" 
			}
		},
		{ &hf_wsp_header_length,
			{ 	"Headers Length",           
				"wsp.headers_length",
				 FT_UINT32, BASE_DEC, NULL, 0x00,
				"Headers Length" 
			}
		},
		{ &hf_wsp_capabilities_section,
			{ 	"Capabilities",           
				"wsp.capabilities",
				 FT_NONE, BASE_DEC, NULL, 0x00,
				"Capabilities" 
			}
		},
		{ &hf_wsp_headers_section,
			{ 	"Headers",           
				"wsp.headers",
				 FT_NONE, BASE_DEC, NULL, 0x00,
				"Headers" 
			}
		},
		{ &hf_wsp_header,
			{ 	"Header",           
				"wsp.headers.header",
				 FT_NONE, BASE_DEC, NULL, 0x00,
				"Header" 
			}
		},
		{ &hf_wsp_header_uri_len,
			{ 	"URI Length",           
				"wsp.uri_length",
				 FT_UINT32, BASE_DEC, NULL, 0x00,
				"URI Length" 
			}
		},
		{ &hf_wsp_header_uri,
			{ 	"URI",           
				"wsp.uri",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"URI" 
			}
		},
		{ &hf_wsp_server_session_id,
			{ 	"Server Session ID",           
				"wsp.server.session_id",
				 FT_UINT32, BASE_DEC, NULL, 0x00,
				"Server Session ID" 
			}
		},
		{ &hf_wsp_header_status,
			{ 	"Status",           
				"wsp.reply.status",
				 FT_UINT8, BASE_HEX, VALS( vals_status ), 0x00,
				"Status" 
			}
		},
		{ &hf_wsp_content_type,
			{ 	"Content Type",           
				"wsp.content_type.type",
				 FT_UINT16, BASE_HEX, VALS ( vals_content_types ), 0x0000,
				"Content Type" 
			}
		},
		{ &hf_wsp_parameter_well_known_charset,
			{ 	"Charset",           
				"wsp.content_type.parameter.charset",
				 FT_UINT16, BASE_HEX, VALS ( vals_character_sets ), 0x00,
				"Charset" 
			}
		},
		{ &hf_wsp_reply_data,
			{ 	"Data",           
				"wsp.reply.data",
				 FT_NONE, BASE_NONE, NULL, 0x00,
				"Data" 
			}
		},
		{ &hf_wsp_header_accept,
			{ 	"Accept",           
				"wsp.header.accept",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_UINT8, BASE_HEX, VALS ( vals_content_types ), 0x00,
				"Accept" 
			}
		},
		{ &hf_wsp_header_accept_str,
			{ 	"Accept",           
				"wsp.header.accept.string",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Accept" 
			}
		},
		{ &hf_wsp_header_accept_charset,
			{ 	"Accept-Charset",           
				"wsp.header.accept_charset",
				 FT_UINT16, BASE_HEX, VALS ( vals_character_sets ), 0x00,
				"Accept-Charset" 
			}
		},
		{ &hf_wsp_header_accept_charset_str,
			{ 	"Accept-Charset",           
				"wsp.header.accept_charset.string",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Accept-Charset" 
			}
		},
		{ &hf_wsp_header_accept_language,
			{ 	"Accept-Language",           
				"wsp.header.accept_language",
				 FT_UINT8, BASE_HEX, VALS ( vals_languages ), 0x00,
				"Accept-Language" 
			}
		},
		{ &hf_wsp_header_accept_language_str,
			{ 	"Accept-Language",           
				"wsp.header.accept_language.string",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Accept-Language" 
			}
		},
		{ &hf_wsp_header_content_language,
			{ 	"Content-Language",           
				"wsp.header.content_language",
				 FT_UINT8, BASE_HEX, VALS ( vals_languages ), 0x00,
				"Content-Language" 
			}
		},
		{ &hf_wsp_header_content_language_str,
			{ 	"Content-Language",           
				"wsp.header.content_language.string",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Content-Language" 
			}
		},
		{ &hf_wsp_header_accept_ranges,
			{ 	"Accept-Ranges",
				"wsp.header.accept_ranges",
				 FT_UINT8, BASE_HEX, VALS ( vals_accept_ranges ), 0x00,
				"Accept-Ranges" 
			}
		},
		{ &hf_wsp_header_accept_ranges_str,
			{ 	"Accept-Ranges",
				"wsp.header.accept_ranges.str",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Accept-Ranges" 
			}
		},
		{ &hf_wsp_header_age,
			{ 	"Age",           
				"wsp.header.age",
				 FT_RELATIVE_TIME, BASE_NONE, NULL, 0x0,
				"Age" 
			}
		},
		{ &hf_wsp_header_cache_control,
			{ 	"Cache-Control",           
				"wsp.header.cache_control",
				 FT_UINT8, BASE_HEX, VALS ( vals_cache_control ), 0x00,
				"Cache-Control" 
			}
		},
		{ &hf_wsp_header_cache_control_tok,
			{ 	"Cache-Control",           
				"wsp.header.cache_control.tok",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Cache-Control" 
			}
		},
		{ &hf_wsp_header_cache_control_header,
			{ 	"Cache-Control: Header",           
				"wsp.header.cache_control.header",
				 FT_UINT8, BASE_HEX, VALS ( vals_header_names ), 0x00,
				"Cache-Control: Header" 
			}
		},
		{ &hf_wsp_header_cache_control_header_tok,
			{ 	"Cache-Control: Header",           
				"wsp.header.cache_control.header_tok",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Cache-Control: Header" 
			}
		},
		{ &hf_wsp_header_cache_control_time,
			{ 	"Cache-Control: time",           
				"wsp.header.cache_control.ttl",
				 FT_RELATIVE_TIME, BASE_NONE, NULL, 0x0,
				"Cache-Control: time" 
			}
		},
		{ &hf_wsp_header_cache_control_max_age,
			{ 	"Cache-Control: Max-Age",           
				"wsp.header.cache_control.max_age",
				 FT_RELATIVE_TIME, BASE_NONE, NULL, 0x0,
				"Cache-Control: Max-Age" 
			}
		},
		{ &hf_wsp_header_cache_control_max_stale,
			{ 	"Cache-Control: Max-Stale",           
				"wsp.header.cache_control.max_stale",
				 FT_RELATIVE_TIME, BASE_NONE, NULL, 0x0,
				"Cache-Control: Max-Stale" 
			}
		},
		{ &hf_wsp_header_cache_control_min_fresh,
			{ 	"Cache-Control: Min-Fresh",           
				"wsp.header.cache_control.min_fresh",
				 FT_RELATIVE_TIME, BASE_NONE, NULL, 0x0,
				"Cache-Control: Min-Fresh" 
			}
		},
		{ &hf_wsp_header_cache_control_no_cache,
			{ 	"Cache-Control: No-Cache",           
				"wsp.header.cache_control.no_cache",
				 FT_UINT8, BASE_HEX, VALS ( vals_header_names ), 0x00,
				"Cache-Control: No-Cache" 
			}
		},
		{ &hf_wsp_header_cache_control_no_cache_str,
			{ 	"Cache-Control: No-Cache",           
				"wsp.header.cache_control.no_cache.str",
				FT_STRING, BASE_NONE, NULL, 0x00,
				"Cache-Control: No-Cache" 
			}
		},
		{ &hf_wsp_header_cache_control_private,
			{ 	"Cache-Control: Private",           
				"wsp.header.cache_control.private",
				 FT_UINT8, BASE_HEX, VALS ( vals_header_names ), 0x00,
				"Cache-Control: Private" 
			}
		},
		{ &hf_wsp_header_cache_control_private_str,
			{ 	"Cache-Control: Private",           
				"wsp.header.cache_control.private.str",
				FT_STRING, BASE_NONE, NULL, 0x00,
				"Cache-Control: Private" 
			}
		},
		{ &hf_wsp_header_content_length,
			{ 	"Content-Length",           
				"wsp.header.content_length",
				 /* FT_UINT32, BASE_DEC, NULL, 0x00, */
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Content-Length" 
			}
		},
		{ &hf_wsp_header_max_forwards,
			{ 	"Max-Forwards",           
				"wsp.header.max_forwards",
				 /* FT_UINT32, BASE_DEC, NULL, 0x00, */
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Max-Forwards" 
			}
		},
		{ &hf_wsp_header_bearer_indication,
			{ 	"Bearer-Indication",           
				"wsp.header.bearer_indication",
				 /* FT_UINT32, BASE_DEC, NULL, 0x00, */
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Bearer-Indication" 
			}
		},
		{ &hf_wsp_header_date,
			{ 	"Date",           
				"wsp.header.date",
				 FT_ABSOLUTE_TIME, BASE_NONE, NULL, 0x0,
				"Date" 
			}
		},
		{ &hf_wsp_header_etag,
			{ 	"Etag",           
				"wsp.header.etag",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Etag" 
			}
		},
		{ &hf_wsp_header_expires,
			{ 	"Expires",           
				"wsp.header.expires",
				 FT_ABSOLUTE_TIME, BASE_NONE, NULL, 0x0,
				"Expires" 
			}
		},
		{ &hf_wsp_header_last_modified,
			{ 	"Last-Modified",           
				"wsp.header.last_modified",
				 FT_ABSOLUTE_TIME, BASE_NONE, NULL, 0x0,
				"Last-Modified" 
			}
		},
		{ &hf_wsp_header_location,
			{ 	"Location",           
				"wsp.header.location",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Location" 
			}
		},
		{ &hf_wsp_header_if_modified_since,
			{ 	"If-Modified-Since",           
				"wsp.header.if_modified_since",
				 FT_ABSOLUTE_TIME, BASE_NONE, NULL, 0x0,
				"If-Modified-Since" 
			}
		},
		{ &hf_wsp_header_server,
			{ 	"Server",           
				"wsp.header.server",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Server" 
			}
		},
		{ &hf_wsp_header_transfer_encoding,
			{ 	"Transfer Encoding",           
				"wsp.header.transfer_enc",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_UINT8, BASE_HEX, VALS ( vals_transfer_encoding ), 0x00,
				"Transfer Encoding" 
			}
		},
		{ &hf_wsp_header_transfer_encoding_str,
			{ 	"Transfer Encoding",           
				"wsp.header.transfer_enc_str",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Transfer Encoding" 
			}
		},
		{ &hf_wsp_header_user_agent,
			{ 	"User-Agent",           
				"wsp.header.user_agent",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"User-Agent" 
			}
		},
		{ &hf_wsp_header_via,
			{ 	"Via",           
				"wsp.header.via",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Via" 
			}
		},
		{ &hf_wsp_header_application_header,
			{ 	"Application Header",           
				"wsp.header.application_header",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Application Header" 
			}
		},
		{ &hf_wsp_header_application_value,
			{ 	"Application Header Value",           
				"wsp.header.application_header.value",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Application Header Value" 
			}
		},
		{ &hf_wsp_header_x_wap_tod,
			{ 	"X-WAP.TOD",           
				"wsp.header.x_wap_tod",
				 FT_ABSOLUTE_TIME, BASE_NONE, NULL, 0x0,
				"X-WAP.TOD" 
			}
		},
		{ &hf_wsp_post_data,
			{ 	"Post Data",           
				"wsp.post.data",
				 FT_NONE, BASE_NONE, NULL, 0x00,
				"Post Data" 
			}
		},
		{ &hf_wsp_header_from,
			{ 	"From",           
				"wsp.header.from",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"From" 
			}
		},
		{ &hf_wsp_header_host,
			{ 	"Host",           
				"wsp.header.host",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Host" 
			}
		},
		{ &hf_wsp_header_if_match,
			{ 	"If-Match",           
				"wsp.header.if_match",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"If-Match" 
			}
		},
		{ &hf_wsp_header_if_none_match,
			{ 	"If-None-Match",           
				"wsp.header.if_none_match",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"If-None-Match" 
			}
		},
		{ &hf_wsp_header_upgrade,
			{ 	"Upgrade",           
				"wsp.header.upgrade",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Upgrade" 
			}
		},
		{ &hf_wsp_header_content_id,
			{ 	"Content-ID",           
				"wsp.header.content_id",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Content-ID" 
			}
		},
		{ &hf_wsp_header_content_base,
			{ 	"Content-Base",           
				"wsp.header.content_base",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Content-Base" 
			}
		},
		{ &hf_wsp_header_content_location,
			{ 	"Content-Location",           
				"wsp.header.content_location",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Content-Location" 
			}
		},
		{ &hf_wsp_header_referer,
			{ 	"Referer",           
				"wsp.header.referer",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Referer" 
			}
		},
		{ &hf_wsp_header_x_wap_content_uri,
			{ 	"X-WAP-Content-URI",           
				"wsp.header.x_wap_content_uri",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"X-WAP-Content-URI" 
			}
		},
		{ &hf_wsp_header_x_wap_initiator_uri,
			{ 	"X-WAP-Initiator-URI",           
				"wsp.header.x_wap_initiator_uri",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"X-WAP-Initiator-URI" 
			}
		},
		{ &hf_wsp_header_profile,
			{ 	"Profile",
				"wsp.header.profile",
				 /*FT_NONE, BASE_DEC, NULL, 0x00,*/
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Profile" 
			}
		},
		{ &hf_wsp_header_profile_warning_code,
			{ 	"Profile-Warning",
				"wsp.header.profile-warning",
				 FT_UINT8, BASE_HEX, VALS ( vals_profile_warning_codes ), 0x00,
				"Profile-Warning"
			}
		},
		{ &hf_wsp_header_profile_warning_uri,
			{ 	"Profile-Warning: URI",
				"wsp.header.profile-warning.uri",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Profile-Warning: URI"
			}
		},
		{ &hf_wsp_header_profile_warning_date,
			{ 	"Profile-Warning: Date",
				"wsp.header.profile-warning.date",
				 FT_ABSOLUTE_TIME, BASE_NONE, NULL, 0x0,
				"Profile-Warning: Date"
			}
		},
		{ &hf_wsp_header_expect_text,
			{ 	"Expect",
				"wsp.header.expect_text",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Expect"
			}
		},
		{ &hf_wsp_header_expect_parm,
			{ 	"Expect Parameter",
				"wsp.header.expect_parm",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Expect Parameter"
			}
		},
		{ &hf_wsp_header_if_unmodified_since,
			{ 	"If-Unmodified-Since",           
				"wsp.header.if_unmodified_since",
				 FT_ABSOLUTE_TIME, BASE_NONE, NULL, 0x0,
				"If-Unmodified-Since" 
			}
		},
		{ &hf_wsp_header_if_range_date,
			{ 	"If-Range",           
				"wsp.header.if_range.date",
				 FT_ABSOLUTE_TIME, BASE_NONE, NULL, 0x0,
				"If-Range" 
			}
		},
		{ &hf_wsp_header_if_range_str,
			{ 	"If-Range",           
				"wsp.header.if_range.str",
				 FT_STRING, BASE_NONE, NULL, 0x0,
				"If-Range" 
			}
		},
		{ &hf_wsp_header_accept_encoding,
			{ 	"Accept-Encoding",           
				"wsp.header.accept_encoding",
				 FT_UINT16, BASE_HEX, VALS ( vals_accept_encodings ), 0x00,
				"Accept-Encoding" 
			}
		},
		{ &hf_wsp_header_accept_encoding_str,
			{ 	"Accept-Encoding",           
				"wsp.header.accept_encoding.string",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Accept-Encoding" 
			}
		},
		{ &hf_wsp_header_content_encoding,
			{ 	"Content-Encoding",           
				"wsp.header.content_encoding",
				 FT_UINT16, BASE_HEX, VALS ( vals_content_encodings ), 0x00,
				"Content-Encoding" 
			}
		},
		{ &hf_wsp_header_content_encoding_str,
			{ 	"Content-Encoding",           
				"wsp.header.content_encoding.string",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Content-Encoding" 
			}
		},
		{ &hf_wsp_header_warning,
			{ 	"Warning",           
				"wsp.header.warning",
				 FT_UINT8, BASE_HEX, VALS ( vals_warnings ), 0x00,
				"Warning" 
			}
		},
		{ &hf_wsp_header_pragma,
			{ 	"Pragma",           
				"wsp.header.pragma",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Pragma" 
			}
		},
		{ &hf_wsp_header_warning_str,
			{ 	"Warning",           
				"wsp.header.warning.string",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Warning" 
			}
		},
		{ &hf_wsp_header_allow,
			{ 	"Allow",           
				"wsp.header.allow",
				 FT_UINT8, BASE_HEX, VALS ( vals_pdu_type ), 0x00,
				"Allow" 
			}
		},
		{ &hf_wsp_header_connection,
			{ 	"Connection",
				"wsp.header.connection",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Connection" 
			}
		},
		{ &hf_wsp_header_public,
			{ 	"Public",
				"wsp.header.public",
				 FT_UINT8, BASE_HEX, VALS ( vals_pdu_type ), 0x00,
				"Public" 
			}
		},
		{ &hf_wsp_header_public_str,
			{ 	"Public",
				"wsp.header.public_str",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Public" 
			}
		},
		{ &hf_wsp_header_vary,
			{ 	"Vary",
				"wsp.header.vary",
				 FT_UINT8, BASE_HEX, VALS ( vals_header_names ), 0x00,
				"Vary" 
			}
		},
		{ &hf_wsp_header_vary_str,
			{ 	"Vary",
				"wsp.header.vary_str",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Vary" 
			}
		},
		{ &hf_wsp_header_trailer,
			{ 	"Trailer",
				"wsp.header.trailer",
				 FT_UINT8, BASE_HEX, VALS ( vals_header_names ), 0x00,
				"Trailer" 
			}
		},
		{ &hf_wsp_header_trailer_str,
			{ 	"Trailer",
				"wsp.header.trailer_str",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Trailer" 
			}
		},
		{ &hf_wsp_header_push_flag,
			{ 	"Push-Flag",
				"wsp.header.push_flag",
				 FT_UINT8, BASE_HEX, NULL, 0x7F,
				"Push-Flag" 
			}
		},
		{ &hf_wsp_header_authorization,
			{ 	"Authorization",
				"wsp.header.authorization",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Authorization" 
			}
		},
		{ &hf_wsp_header_proxy_authorization,
			{ 	"Proxy-Authorization",
				"wsp.header.proxy_authorization",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Proxy-Authorization" 
			}
		},
		{ &hf_wsp_header_www_authenticate,
			{ 	"WWW-Authenticate",
				"wsp.header.www_authenticate",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"WWW-Authenticate" 
			}
		},
		{ &hf_wsp_header_proxy_authenticate,
			{ 	"Proxy-Authenticate",
				"wsp.header.proxy_authenticate",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Proxy-Authenticate" 
			}
		},
		/* Redirect PDU */
		{ &hf_wsp_redirect_addresses_section,
			{ 	"Redirect addresses",
				"wsp.redirect.addresses",
				 FT_NONE, BASE_DEC, NULL, 0x00,
			 	"Redirect addresses"
			}
		},
		{ &hf_wsp_redirect_address_section,
			{ 	"Redirect address record",
				"wsp.redirect.addresses.address_record",
				 FT_NONE, BASE_DEC, NULL, 0x00,
			 	"Redirect address record"
			}
		},
		{ &hf_wsp_redirect_flags,
			{	"Redirect flags",
				"wsp.redirect.flags",
				FT_UINT8, BASE_HEX, VALS( vals_wsp_redirect_flags ), 0xFF,
				"Redirect flags"
			}
		},
		{ &hf_wsp_redirect_bearertype_included,
			{ 	"Bearer Type included",
				"wsp.redirect.addresses.has_bearertype",
				 FT_BOOLEAN, 8, NULL, 0x80,
			 	 "Bearer Type included",
			}
		},
		{ &hf_wsp_redirect_bearertype,
			{ 	"Bearer Type",
				"wsp.redirect.addresses.address.bearertype",
				 FT_UINT8, BASE_DEC, NULL, 0x00,
			 	 "Bearer Type",
			}
		},
		{ &hf_wsp_redirect_port_included,
			{ 	"Port Number included",
				"wsp.redirect.addresses.has_portnumber",
				 FT_BOOLEAN, 8, NULL, 0x40,
			 	 "Port Number included",
			}
		},
		{ &hf_wsp_redirect_port,
			{ 	"Port Number",
				"wsp.redirect.addresses.address.portnumber",
				 FT_UINT16, BASE_DEC, NULL, 0x00,
			 	 "Port Number",
			}
		},
		{ &hf_wsp_redirect_address_length,
			{ 	"Address length",
				"wsp.redirect_addresses.address.length",
				 FT_UINT8, BASE_DEC, NULL, 0x3F,
			 	"Redirect addresses"
			}
		},
		{ &hf_wsp_capability_section,
			{ 	"Capability",           
				"wsp.capability",
				 FT_NONE, BASE_DEC, NULL, 0x00,
				"Capability" 
			}
		},
		{ &hf_wsp_capability_client_sdu_size,
			{ 	"Client SDU size",
				"wsp.capability.client-sdu-size",
				 FT_UINT32, BASE_DEC, NULL, 0x00,
				"Client SDU size" 
			}
		},
		{ &hf_wsp_capability_server_sdu_size,
			{ 	"Server SDU size",
				"wsp.capability.server-sdu-size",
				 FT_UINT32, BASE_DEC, NULL, 0x00,
				"Server SDU size" 
			}
		},
		{ &hf_wsp_capability_method_mor,
			{ 	"Maximum Outstanding Requests for Methods",           
				"wsp.capability.method-mor",
				 FT_UINT8, BASE_DEC, NULL, 0x00,
				"Maximum Outstanding Requests for Methods" 
			}
		},
		{ &hf_wsp_capability_push_mor,
			{ 	"Maximum Outstanding Requests for Push",           
				"wsp.capability.push-mor",
				 FT_UINT8, BASE_DEC, NULL, 0x00,
				"Maximum Outstanding Requests for Push" 
			}
		},
		{ &hf_wsp_capability_proto_options,
			{ 	"Protocol options",
				"wsp.capability.protocol_options",
				 FT_NONE, BASE_HEX, NULL, 0x00,
				"Protocol options" 
			}
		},
		{ &hf_wsp_capability_proto_option,
			{ 	"Protocol option",
				"wsp.capability.protocol_option",
				 FT_UINT8, BASE_HEX, VALS ( vals_wsp_capability_proto_options ), 0xFF,
				"Protocol option" 
			}
		},
		{ &hf_wsp_capability_xmethods,
			{ 	"Extended methods",
				"wsp.capability.xmethods",
				 FT_NONE, BASE_HEX, NULL, 0x00,
				"Extended methods" 
			}
		},
		{ &hf_wsp_capability_xmethod_get,
			{ 	"Extended GET method",
				"wsp.capability.xmethod.get",
				 FT_UINT8, BASE_HEX, NULL, 0x00,
				"Extended GET method" 
			}
		},
		{ &hf_wsp_capability_xmethod_get_val,
			{ 	"Extended GET method value",
				"wsp.capability.xmethod.get-val",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Extended GET method value" 
			}
		},
		{ &hf_wsp_capability_xmethod_push,
			{ 	"Extended PUSH method",
				"wsp.capability.xmethod.push",
				 FT_UINT8, BASE_HEX, NULL, 0x00,
				"Extended PUSH method" 
			}
		},
		{ &hf_wsp_capability_xmethod_push_val,
			{ 	"Extended PUSH method value",
				"wsp.capability.xmethod.push-val",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Extended PUSH method value" 
			}
		},
		{ &hf_wsp_capability_header_code_pages,
			{ 	"Header code pages",
				"wsp.capability.header-code-pages",
				 FT_NONE, BASE_HEX, NULL, 0x00,
				"Header code pages" 
			}
		},
		{ &hf_wsp_capability_header_code_page,
			{ 	"Header code page",
				"wsp.capability.xmethod.hcp",
				 FT_UINT8, BASE_HEX, NULL, 0x00,
				"Header code page" 
			}
		},
		{ &hf_wsp_capability_header_code_page_val,
			{ 	"Header code page value",
				"wsp.capability.xmethod.hcp-val",
				 FT_STRING, BASE_NONE, NULL, 0x00,
				"Header code page value" 
			}
		},
		{ &hf_wsp_capability_alias_address_section,
			{ 	"Alias address record",
				"wsp.capability.alias.addresses.address_record",
				 FT_NONE, BASE_DEC, NULL, 0x00,
			 	"Redirect address record"
			}
		},
		{ &hf_wsp_capability_alias_bearertype_included,
			{ 	"Bearer Type included",
				"wsp.capability.alias.addresses.has_bearertype",
				 FT_BOOLEAN, 8, NULL, 0x80,
			 	"Bearer Type included",
			}
		},
		{ &hf_wsp_capability_alias_bearertype,
			{ 	"Bearer Type",
				"wsp.capability.alias.addresses.address.bearertype",
				 FT_UINT8, BASE_DEC, NULL, 0x00,
			 	"Bearer Type",
			}
		},
		{ &hf_wsp_capability_alias_port_included,
			{ 	"Port Number included",
				"wsp.capability.alias.addresses.has_portnumber",
				 FT_BOOLEAN, 8, NULL, 0x40,
			 	"Port Number included",
			}
		},
		{ &hf_wsp_capability_alias_port,
			{ 	"Port Number",
				"wsp.capability.alias.addresses.address.portnumber",
				 FT_UINT16, BASE_DEC, NULL, 0x00,
			 	"Port Number",
			}
		},
		{ &hf_wsp_capability_alias_address_length,
			{ 	"Address length",
				"wsp.capability.alias.address.length",
				 FT_UINT8, BASE_DEC, NULL, 0x3F,
			 	"Address length"
			}
		},
	};


/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_wsp,
		&ett_header,
		&ett_headers,
		&ett_capabilities,
		&ett_content_type,
		&ett_redirect,
		&ett_redir_addr,
	};

/* Register the protocol name and description */
	proto_wsp = proto_register_protocol(
		"Wireless Session Protocol",   	/* protocol name for use by ethereal */ 
		"WSP",                          /* short version of name */
		"wap-wsp"                    	/* Abbreviated protocol name, should Match IANA 
						    < URL:http://www.isi.edu/in-notes/iana/assignments/port-numbers/ >
						  */
	);

/* Required function calls to register the header fields and subtrees used  */
	proto_register_field_array(proto_wsp, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	register_dissector("wsp", dissect_wsp, proto_wsp);
};

void
proto_reg_handoff_wsp(void)
{
	/*
	 * Get a handle for the WMLC dissector
	 */
	wmlc_handle = find_dissector("wmlc");	/* Coming soon :) */

	/* Only connection-less WSP has no previous handler */
	dissector_add("udp.port", UDP_PORT_WSP, dissect_wsp, proto_wsp);

	/* This dissector is also called from the WTP and WTLS dissectors */
}
